from cpython cimport PyObject
from cpython.ref cimport Py_DECREF
cimport cython
from pygame._sdl2.sdl2 import error
from pygame._sdl2.sdl2 import error as errorfnc
from libc.stdlib cimport free, malloc


WINDOWPOS_UNDEFINED = _SDL_WINDOWPOS_UNDEFINED
WINDOWPOS_CENTERED = _SDL_WINDOWPOS_CENTERED

MESSAGEBOX_ERROR = _SDL_MESSAGEBOX_ERROR
MESSAGEBOX_WARNING = _SDL_MESSAGEBOX_WARNING
MESSAGEBOX_INFORMATION = _SDL_MESSAGEBOX_INFORMATION

SCALEQUALITY_NEAREST=SDL_ScaleMode.SDL_ScaleModeNearest
SCALEQUALITY_LINEAR=SDL_ScaleMode.SDL_ScaleModeLinear
SCALEQUALITY_BEST=SDL_ScaleMode.SDL_ScaleModeBest

import_pygame_base()
import_pygame_color()
import_pygame_surface()
import_pygame_rect()
import_pygame_window()

class RendererDriverInfo:
    def __repr__(self):
        return "<%s(name: %s, flags: 0x%02x, num_texture_formats: %d, max_texture_width: %d, max_texture_height: %d)>" % (
            self.__class__.__name__,
            self.name,
            self.flags,
            self.num_texture_formats,
            self.max_texture_width,
            self.max_texture_height,
        )

def get_drivers():
    """Yield info about the rendering drivers available for Renderer objects
    """
    cdef int num = SDL_GetNumRenderDrivers()
    cdef SDL_RendererInfo info
    cdef int ind
    for ind from 0 <= ind < num:
        SDL_GetRenderDriverInfo(ind, &info)
        ret = RendererDriverInfo()
        ret.name = info.name.decode("utf-8")
        ret.flags = info.flags
        ret.num_texture_formats = info.num_texture_formats
        ret.max_texture_width = info.max_texture_width
        ret.max_texture_height = info.max_texture_height
        yield ret


def get_grabbed_window():
    """Get the window with input grab enabled

    Gets the :class:`Window` with input grab enabled, if input is grabbed,
    otherwise ``None`` is returned.
    """
    cdef SDL_Window *win = SDL_GetGrabbedWindow()
    cdef void *ptr
    if win:
        ptr = SDL_GetWindowData(win, "pg_window")
        if not ptr:
            return None
        return <object>ptr
    return None

def messagebox(title, message,
               Window window=None,
               bint info=False,
               bint warn=False,
               bint error=False,
               buttons=('OK', ),
               return_button=0,
               escape_button=0):
    """Create a native GUI message box

    Creates a native GUI message box.

    :param str title: A title string, or ``None`` to omit a title.
    :param str message: A message string.
    :param bool info: If ``True``, display an info message.
    :param bool warn: If ``True``, display a warning message.
    :param bool error: If ``True``, display an error message.
    :param tuple buttons: An optional sequence of button name strings to show to the user.
    :param int return_button: Button index to use if the return key is hit (``-1`` for none).
    :param int escape_button: Button index to use if the escape key is hit (``-1`` for none).
    :return: The index of the button that was pushed.
    """
    # TODO: type check
    # TODO: color scheme
    cdef SDL_MessageBoxButtonData* c_buttons = NULL

    cdef SDL_MessageBoxData data
    data.flags = 0
    if warn:
        data.flags |= _SDL_MESSAGEBOX_WARNING
    if error:
        data.flags |= _SDL_MESSAGEBOX_ERROR
    if info:
        data.flags |= _SDL_MESSAGEBOX_INFORMATION
    if not window:
        data.window = NULL
    else:
        data.window = window._win
    if title is not None:
        title = title.encode('utf8')
        data.title = title
    else:
        data.title = NULL
    message = message.encode('utf8')
    data.message = message
    data.colorScheme = NULL

    cdef SDL_MessageBoxButtonData button
    if not buttons:
        button.flags = _SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT |\
                       _SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT
        button.buttonid = 0
        button.text = "OK"
        data.buttons = &button
        data.numbuttons = 1
    else:
        buttons_utf8 = [s.encode('utf8') for s in buttons]
        data.numbuttons = <int>len(buttons)
        c_buttons =\
            <SDL_MessageBoxButtonData*>malloc(data.numbuttons * sizeof(SDL_MessageBoxButtonData))
        if not c_buttons:
            raise MemoryError()
        for i, but in enumerate(reversed(buttons_utf8)):
            c_buttons[i].flags = 0
            c_buttons[i].buttonid = data.numbuttons - i - 1
            if c_buttons[i].buttonid == return_button:
                c_buttons[i].flags |= _SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT
            if c_buttons[i].buttonid == escape_button:
                c_buttons[i].flags |= _SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT
            c_buttons[i].text = but
        data.buttons = c_buttons

    cdef int buttonid
    if SDL_ShowMessageBox(&data, &buttonid):
        free(c_buttons)
        raise errorfnc()

    free(c_buttons)
    return buttonid

globals()["Window"] = Window
_Window = Window

cdef Uint32 format_from_depth(int depth):
    cdef Uint32 Rmask, Gmask, Bmask, Amask
    if depth == 16:
        Rmask = 0xF << 8
        Gmask = 0xF << 4
        Bmask = 0xF
        Amask = 0xF << 12
    elif depth in (0, 32):
        Rmask = 0xFF << 16
        Gmask = 0xFF << 8
        Bmask = 0xFF
        Amask = 0xFF << 24
    else:
        raise ValueError("no standard masks exist for given bitdepth with alpha")
    return SDL_MasksToPixelFormatEnum(depth,
                                      Rmask, Gmask, Bmask, Amask)


# disable auto_pickle since it causes stubcheck error
@cython.auto_pickle(False)
cdef class Texture:

    def __init__(self,
                 Renderer renderer,
                 size, int depth=0,
                 static=False, streaming=False,
                 target=False, scale_quality=None):
        """pygame object that represents a texture

        Creates an empty texture.

        :param Renderer renderer: The rendering context for the texture.
        :param tuple size: The width and height for the texture.
        :param int depth: The pixel format (``0`` to use the default).
        :param bool static: Initialize the texture as static
                            (changes rarely, not lockable).
        :param bool streaming: Initialize the texture as streaming
                               (changes frequently, lockable).
        :param bool target: Initialize the texture as target
                            (can be used as a rendering target).
        :param int scale_quality: Set the texture scale quality. Can be ``0``
                                  (nearest pixel sampling), ``1``
                                  (linear filtering, supported by OpenGL and Direct3D)
                                  and ``2`` (anisotropic filtering, supported by Direct3D).

        One of ``static``, ``streaming``, or ``target`` can be set
        to ``True``. If all are ``False``, then ``static`` is used.


        :class:`Texture` objects provide a platform-agnostic API for working with GPU textures.
        They are stored in GPU video memory (VRAM), and are therefore very fast to
        rotate and resize when drawn onto a :class:`Renderer`
        (an object that manages a rendering context inside a :class:`Window`) on most GPUs.

        Since textures are stored in GPU video memory, they aren't as easy to modify
        as the image data of :class:`pygame.Surface` objects, which reside in RAM.

        Textures can be modified in 2 ways:

           * By drawing other textures onto them, achieved by marking them as "target" textures and setting them as the rendering target of their Renderer object (if properly configured and supported).

           * By updating them with a Surface.

            .. note::
                A :class:`pygame.Surface`-to-:class:`Texture` update is generally
                considered a slow operation, as it requires image data to be uploaded from
                RAM to VRAM, which can have a notable overhead cost.
        """
        # https://wiki.libsdl.org/SDL_CreateTexture
        # TODO: masks
        cdef Uint32 format
        try:
            format = format_from_depth(depth)
        except ValueError as e:
            raise e

        cdef int width, height
        if len(size) != 2:
            raise ValueError('size must have two elements')
        width, height = size[0], size[1]
        if width <= 0 or height <= 0:
            raise ValueError('size must contain two positive values')

        cdef int access
        if static:
            if streaming or target:
                raise ValueError('only one of static, streaming, or target can be true')
            access = _SDL_TEXTUREACCESS_STATIC
        elif streaming:
            if static or target:
                raise ValueError('only one of static, streaming, or target can be true')
            access = _SDL_TEXTUREACCESS_STREAMING
        elif target:
            if streaming or static:
                raise ValueError('only one of static, streaming, or target can be true')
            access = _SDL_TEXTUREACCESS_TARGET
        else:
            # Create static texture by default.
            access = _SDL_TEXTUREACCESS_STATIC

        self.renderer = renderer
        cdef SDL_Renderer* _renderer = renderer._renderer
        self._tex = SDL_CreateTexture(_renderer,
                                      format,
                                      access,
                                      width, height)
        if not self._tex:
            raise error()

        if not scale_quality is None:
            if SDL_VERSION_ATLEAST(2,0,12):
                SDL_SetTextureScaleMode(self._tex,scale_quality)
            else:
                SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY,{
                    0: b'nearest',
                    1: b'linear',
                    2: b'best'
                }[scale_quality])

        self.width, self.height = width, height

    @staticmethod
    def from_surface(Renderer renderer, surface):
        """Create a texture from an existing surface

        :param Renderer renderer: Rendering context for the texture.
        :param pygame.Surface surface: The surface to create a texture from.
        """
        # https://wiki.libsdl.org/SDL_CreateTextureFromSurface
        if not pgSurface_Check(surface):
            raise TypeError('2nd argument must be a surface')
        cdef Texture self = Texture.__new__(Texture)
        self.renderer = renderer
        cdef SDL_Renderer* _renderer = renderer._renderer
        cdef SDL_Surface *surf_ptr = pgSurface_AsSurface(surface)
        self._tex = SDL_CreateTextureFromSurface(_renderer,
                                                 surf_ptr)
        if not self._tex:
            raise error()
        self.width = surface.get_width()
        self.height = surface.get_height()
        return self

    def __dealloc__(self):
        if self._tex:
            SDL_DestroyTexture(self._tex)

    @property
    def alpha(self):
        """Get or set the additional alpha value multiplied into draw operations
        """
        # https://wiki.libsdl.org/SDL_GetTextureAlphaMod
        cdef Uint8 alpha
        cdef int res = SDL_GetTextureAlphaMod(self._tex, &alpha)
        if res < 0:
            raise error()

        return alpha

    @alpha.setter
    def alpha(self, Uint8 new_value):
        # https://wiki.libsdl.org/SDL_SetTextureAlphaMod
        cdef int res = SDL_SetTextureAlphaMod(self._tex, new_value)
        if res < 0:
            raise error()

    @property
    def blend_mode(self):
        """Get or set the blend mode for texture drawing operations

        Gets or sets the blend mode for the texture's drawing operations.
        Valid blend modes are any of the ``BLENDMODE_*`` constants or a custom one.
        """
        # https://wiki.libsdl.org/SDL_GetTextureBlendMode
        cdef SDL_BlendMode blendMode
        cdef int res = SDL_GetTextureBlendMode(self._tex, &blendMode)
        if res < 0:
            raise error()

        return blendMode

    @blend_mode.setter
    def blend_mode(self, blendMode):
        # https://wiki.libsdl.org/SDL_SetTextureBlendMode
        cdef int res = SDL_SetTextureBlendMode(self._tex, blendMode)
        if res < 0:
            raise error()

    @property
    def color(self):
        """Get or set the additional color value multiplied into texture drawing operations
        """
        cdef Uint8[4] rgba

        # https://wiki.libsdl.org/SDL_GetTextureColorMod
        cdef int res = SDL_GetTextureColorMod(self._tex,
                                              &(rgba[0]),
                                              &(rgba[1]),
                                              &(rgba[2]))
        rgba[3] = 255

        if res < 0:
            raise error()

        return pgColor_NewLength(rgba, 4)

    @color.setter
    def color(self, new_value):
        cdef Uint8[4] rgba
        pg_RGBAFromObjEx(new_value, rgba, PG_COLOR_HANDLE_ALL)

        # https://wiki.libsdl.org/SDL_SetTextureColorMod
        cdef int res = SDL_SetTextureColorMod(self._tex,
                                              rgba[0],
                                              rgba[1],
                                              rgba[2])

        if res < 0:
            raise error()

    def get_rect(self, **kwargs):
        """Get the rectangular area of the texture

        Like :meth:`pygame.Surface.get_rect`, this method returns a **new**
        :class:`pygame.Rect` covering the entire texture. It will always
        have a ``topleft`` of ``(0, 0)`` with a ``width`` and ``height`` the same size as the
        texture.
        """
        rect = pgRect_New4(0, 0, self.width, self.height)
        for key in kwargs:
            setattr(rect, key, kwargs[key])

        return rect

    cpdef void draw(self, srcrect=None, dstrect=None, float angle=0, origin=None,
                    bint flip_x=False, bint flip_y=False):
        """Copy a portion of the texture to the rendering target

        :param srcrect: The source rectangle on the texture, or ``None`` for the
                        entire texture.
        :param dstrect: The destination rectangle on the rendering target, or
                        ``None`` for the entire rendering target. The texture
                        will be stretched to fill ``dstrect``.
        :param float angle: The angle (in degrees) to rotate dstrect around
                            (clockwise).
        :param origin: The point around which dstrect will be rotated.
                       If ``None``, it will equal the center:
                       ``(dstrect.w/2, dstrect.h/2)``.
        :param bool flip_x: Flip the drawn texture portion horizontally (x - axis).
        :param bool flip_y: Flip the drawn texture portion vertically (y - axis).
        """
        cdef SDL_Rect src
        cdef SDL_Rect *csrcrect = NULL
        cdef SDL_FRect dst
        cdef SDL_FRect *cdstrect = NULL
        cdef SDL_FPoint corigin
        cdef SDL_FPoint *originptr

        if srcrect is not None:
            csrcrect = pgRect_FromObject(srcrect, &src)
            if not csrcrect:
                raise TypeError("the argument is not a rectangle or None")

        if dstrect is not None:
            cdstrect = pgFRect_FromObject(dstrect, &dst)
            if cdstrect == NULL:
                if len(dstrect) == 2:
                    dst.x = dstrect[0]
                    dst.y = dstrect[1]
                    dst.w = <float> self.width
                    dst.h = <float> self.height
                    cdstrect = &dst
                else:
                    raise TypeError('dstrect must be a position, rect, or None')

        if origin:
            originptr = &corigin
            corigin.x = origin[0]
            corigin.y = origin[1]
        else:
            originptr = NULL

        cdef int flip = SDL_FLIP_NONE
        if flip_x:
            flip |= SDL_FLIP_HORIZONTAL
        if flip_y:
            flip |= SDL_FLIP_VERTICAL

        cdef int res = SDL_RenderCopyExF(self.renderer._renderer, self._tex, csrcrect, cdstrect,
                                        angle, originptr, <SDL_RendererFlip>flip)
        if res < 0:
            raise error()

    def draw_triangle(self, p1_xy, p2_xy, p3_xy,
                      p1_uv=(0.0, 0.0), p2_uv=(1.0, 1.0), p3_uv=(0.0, 1.0),
                      p1_mod=(255, 255, 255, 255), p2_mod=(255, 255, 255, 255), p3_mod=(255, 255, 255, 255)):
        """Copy a triangle portion of the texture to the rendering target using the given coordinates

        :param p1_xy: The first vertex coordinate on the rendering target.
        :param p2_xy: The second vertex coordinate on the rendering target.
        :param p3_xy: The third vertex coordinate on the rendering target.
        :param p1_uv: The first UV coordinate of the triangle portion.
        :param p2_uv: The second UV coordinate of the triangle portion.
        :param p3_uv: The third UV coordinate of the triangle portion.
        :param p1_mod: The first vertex color modulation.
        :param p2_mod: The second vertex color modulation.
        :param p3_mod: The third vertex color modulation.
        """
        if not SDL_VERSION_ATLEAST(2, 0, 18):
            raise error("draw_triangle requires SDL 2.0.18 or newer")

        cdef Uint8 _r_mod, _g_mod, _b_mod, _a_mod
        SDL_GetTextureColorMod(self._tex, &_r_mod, &_g_mod, &_b_mod)
        SDL_GetTextureAlphaMod(self._tex, &_a_mod)

        cdef float r_mod = <float>_r_mod / <float>255.0
        cdef float g_mod = <float>_g_mod / <float>255.0
        cdef float b_mod = <float>_b_mod / <float>255.0
        cdef float a_mod = <float>_a_mod / <float>255.0

        cdef SDL_Vertex vertices[3]
        for i, vert in enumerate(((p1_xy, p1_mod, p1_uv),
                                  (p2_xy, p2_mod, p2_uv),
                                  (p3_xy, p3_mod, p3_uv))):
            xy, mod, uv = vert
            vertices[i].position.x = xy[0]
            vertices[i].position.y = xy[1]
            vertices[i].color.r = r_mod * mod[0]
            vertices[i].color.g = g_mod * mod[1]
            vertices[i].color.b = b_mod * mod[2]
            vertices[i].color.a = a_mod * mod[3] if len(mod) > 3 else _a_mod
            vertices[i].tex_coord.x = uv[0]
            vertices[i].tex_coord.y = uv[1]

        cdef int res = SDL_RenderGeometry(self.renderer._renderer, self._tex, vertices, 3, NULL, 0)
        if res < 0:
            raise error()

    def draw_quad(self, p1_xy, p2_xy, p3_xy, p4_xy,
                  p1_uv=(0.0, 0.0), p2_uv=(1.0, 0.0),
                  p3_uv=(1.0, 1.0), p4_uv=(0.0, 1.0),
                  p1_mod=(255, 255, 255, 255), p2_mod=(255, 255, 255, 255),
                  p3_mod=(255, 255, 255, 255), p4_mod=(255, 255, 255, 255)):
        """Copy a quad portion of the texture to the rendering target using the given coordinates

        :param p1_xy: The first vertex coordinate on the rendering target.
        :param p2_xy: The second vertex coordinate on the rendering target.
        :param p3_xy: The third vertex coordinate on the rendering target.
        :param p4_xy: The fourth vertex coordinate on the rendering target.
        :param p1_uv: The first UV coordinate of the quad portion.
        :param p2_uv: The second UV coordinate of the quad portion.
        :param p3_uv: The third UV coordinate of the quad portion.
        :param p4_uv: The fourth UV coordinate of the quad portion.
        :param p1_mod: The first vertex color modulation.
        :param p2_mod: The second vertex color modulation.
        :param p3_mod: The third vertex color modulation.
        :param p4_mod: The fourth vertex color modulation.
        """
        if not SDL_VERSION_ATLEAST(2, 0, 18):
            raise error("draw_quad requires SDL 2.0.18 or newer")

        cdef Uint8 _r_mod, _g_mod, _b_mod, _a_mod
        SDL_GetTextureColorMod(self._tex, &_r_mod, &_g_mod, &_b_mod)
        SDL_GetTextureAlphaMod(self._tex, &_a_mod)

        cdef float r_mod = <float>_r_mod / <float>255.0
        cdef float g_mod = <float>_g_mod / <float>255.0
        cdef float b_mod = <float>_b_mod / <float>255.0
        cdef float a_mod = <float>_a_mod / <float>255.0

        cdef SDL_Vertex vertices[6]
        for i, vert in enumerate(((p1_xy, p1_mod, p1_uv),
                                  (p2_xy, p2_mod, p2_uv),
                                  (p3_xy, p3_mod, p3_uv),
                                  (p3_xy, p3_mod, p3_uv),
                                  (p4_xy, p4_mod, p4_uv),
                                  (p1_xy, p1_mod, p1_uv))):
            xy, mod, uv = vert
            vertices[i].position.x = xy[0]
            vertices[i].position.y = xy[1]
            vertices[i].color.r = r_mod * mod[0]
            vertices[i].color.g = g_mod * mod[1]
            vertices[i].color.b = b_mod * mod[2]
            vertices[i].color.a = a_mod * mod[3] if len(mod) > 3 else _a_mod
            vertices[i].tex_coord.x = uv[0]
            vertices[i].tex_coord.y = uv[1]

        cdef int res = SDL_RenderGeometry(self.renderer._renderer, self._tex, vertices, 6, NULL, 0)
        if res < 0:
            raise error()

    def update(self, surface, area=None):
        # https://wiki.libsdl.org/SDL_UpdateTexture
        # Should it accept a raw pixel data array too?
        """Update the texture with Surface (slow operation, use sparingly)

        :param Surface surface: The source surface.
        :param area: The rectangular area of the texture to update.

        .. note::
            This is a fairly slow function, intended for use with static textures
            that do not change often.

            If the texture is intended to be updated often, it is preferred to create
            the texture as streaming and use the locking functions.

            While this function will work with streaming textures, for optimization
            reasons you may not get the pixels back if you lock the texture afterward.
        """

        if not pgSurface_Check(surface):
            raise TypeError("update source should be a Surface.")

        cdef SDL_Rect rect
        rect.x = 0
        rect.y = 0
        cdef SDL_Rect *rectptr = pgRect_FromObject(area, &rect)
        cdef SDL_Surface *surf = pgSurface_AsSurface(surface)

        if rectptr == NULL and area is not None:
            raise TypeError('area must be a rectangle or None')

        cdef int dst_width, dst_height
        if rectptr == NULL:
            dst_width = self.width
            dst_height = self.height
        else:
            dst_width = rect.w
            dst_height = rect.h

        if dst_height > surf.h or dst_width > surf.w:
            # if the surface is smaller than the destination rect,
            # clip the rect to prevent segfault
            rectptr = &rect # make sure rectptr is not NULL
            rect.h = surf.h
            rect.w = surf.w

        # For converting the surface, if needed
        cdef SDL_Surface *converted_surf = NULL
        cdef SDL_PixelFormat *pixel_format = NULL
        cdef SDL_BlendMode blend

        cdef Uint32 format_
        if (SDL_QueryTexture(self._tex, &format_, NULL, NULL, NULL) != 0):
            raise error()

        cdef int res
        if format_ != surf.format.format:
            if (SDL_GetSurfaceBlendMode(surf, &blend) != 0):
                raise error()

            pixel_format = SDL_AllocFormat(format_)
            if (pixel_format == NULL):
                raise error()

            converted_surf = SDL_ConvertSurface(surf, pixel_format, 0)
            if (SDL_SetSurfaceBlendMode(converted_surf, blend) != 0):
                SDL_FreeSurface(converted_surf)
                SDL_FreeFormat(pixel_format)
                raise error()

            res = SDL_UpdateTexture(self._tex, rectptr, converted_surf.pixels, converted_surf.pitch)
            SDL_FreeSurface(converted_surf)
            SDL_FreeFormat(pixel_format)
        else:
            res = SDL_UpdateTexture(self._tex, rectptr, surf.pixels, surf.pitch)

        if res < 0:
            raise error()

# disable auto_pickle since it causes stubcheck error
@cython.auto_pickle(False)
cdef class Image:

    def __cinit__(self):
        self.angle = 0
        self._origin.x = 0
        self._origin.y = 0
        self._originptr = NULL
        self.flip_x = False
        self.flip_y = False

        cdef Uint8[4] defaultColor = [255, 255, 255, 255]
        self._color = pgColor_NewLength(defaultColor, 4)
        self.alpha = 255

    def __init__(self, texture_or_image, srcrect=None):
        """pygame object that represents a portion of a texture

        Creates an Image.

        :param Texture | Image texture_or_image: The Texture or an existing Image
                                                 object to create the Image from.
        :param srcrect: The rectangular portion of the Texture or Image object
                        passed to ``texture_or_image``.

        An :class:`Image` object represents a portion of a :class:`Texture`. Specifically,
        they can be used to store and manipulate arguments for :meth:`Texture.draw`
        in a more user-friendly way.
        """
        cdef SDL_Rect temp
        cdef SDL_Rect *rectptr

        if isinstance(texture_or_image, Image):
            self.texture = texture_or_image.texture
            self.srcrect = pgRect_New(&(<Rect>texture_or_image.srcrect).r)
        else:
            self.texture = texture_or_image
            self.srcrect = texture_or_image.get_rect()
        self.blend_mode = texture_or_image.blend_mode

        if srcrect is not None:
            rectptr = pgRect_FromObject(srcrect, &temp)
            if rectptr == NULL:
                raise TypeError('srcrect must be None or a rectangle')
            temp.x = rectptr.x
            temp.y = rectptr.y
            temp.w = rectptr.w
            temp.h = rectptr.h

            if temp.x < 0 or temp.y < 0 or \
                temp.w < 0 or temp.h < 0 or \
                temp.x + temp.w > self.srcrect.w or \
                temp.y + temp.h > self.srcrect.h:
                raise ValueError('rect values are out of range')
            temp.x += self.srcrect.x
            temp.y += self.srcrect.y
            self.srcrect = pgRect_New(&temp)

    @property
    def color(self):
        """Gets or sets the Image color modifier
        """
        return self._color

    @color.setter
    def color(self, new_color):
        cdef Uint8[4] rgba
        pg_RGBAFromObjEx(new_color, rgba, PG_COLOR_HANDLE_ALL)

        self._color[:3] = rgba[:3]

    @property
    def origin(self):
        """Get or set the Image's origin of rotation

        Gets or sets the Image's origin of rotation, a 2D coordinate relative to the topleft coordinate of the Image's rectangular area.

        An origin of ``None`` means no origin was set and the Image will be rotated around
        its center.
        """
        if self._originptr == NULL:
            return None
        else:
            return (self._origin.x, self._origin.y)

    @origin.setter
    def origin(self, new_origin):
        if new_origin:
            self._origin.x = <int>new_origin[0]
            self._origin.y = <int>new_origin[1]
            self._originptr = &self._origin
        else:
            self._originptr = NULL

    def get_rect(self):
        """Get the rectangular area of the Image

        .. note::
            The returned :class:`Rect` object might contain position information
            relative to the bounds of the :class:`Texture` or Image object it was
            created from.
        """
        return pgRect_New(&self.srcrect.r)

    cpdef void draw(self, srcrect=None, dstrect=None):
        """Copy a portion of the Image to the rendering target

        :param srcrect: Source rectangle specifying a sub-Image, or None for the
                        entire Image.
        :param dstrect: Destination rectangle or position on the render target,
                        or ``None`` for entire target. The Image is stretched to
                        fill dstrect.
        """

        self.texture.color = self._color
        self.texture.alpha = self.alpha
        self.texture.blend_mode = self.blend_mode
        self.texture.draw(
            (srcrect if not srcrect is None else self.srcrect),
            dstrect,
            self.angle,
            self.origin,
            self.flip_x,
            self.flip_y)

# disable auto_pickle since it causes stubcheck error
@cython.auto_pickle(False)
cdef class Renderer:

    @classmethod
    def from_window(cls, Window window):
        cdef Renderer self = cls.__new__(cls)
        self._win = window
        if self._win._is_borrowed:
            self._is_borrowed=1
        else:
            raise error()
        if not self._win:
            raise error()

        self._renderer =  SDL_GetRenderer(self._win._win)
        if not self._renderer:
            raise error()

        self._target = None
        return self

    def __init__(self,Window window, int index=-1,
                 int accelerated=-1, bint vsync=False,
                 bint target_texture=False):
        """pygame object wrapping a 2D rendering context for a window

        Creates a 2D rendering context for a window.

        :param Window window: The window onto which the rendering context should be
                              placed.
        :param int index: The index of rendering driver to initialize, or ``-1`` to init
                          the first supporting the requested options.
        :param int accelerated: If 1, the renderer uses hardware acceleration.
                                if 0, the renderer is a software fallback.
                                ``-1`` gives precedence to renderers using hardware
                                acceleration.
        :param bool vsync: If ``True`` :meth:`Renderer.present` is synchronized with
                           the refresh rate.
        :param bool target_texture: Whether the renderer should support setting
                                   :class:`Texture` objects as target textures, to
                                   enable drawing onto them.


        :class:`Renderer` objects provide a cross-platform API for rendering 2D
        graphics onto a :class:`Window`, by using either Metal (macOS), OpenGL
        (macOS, Windows, Linux) or Direct3D (Windows) rendering drivers, depending
        on what is set or is available on a system during their creation.

        They can be used to draw both :class:`Texture` objects and simple points,
        lines and rectangles (which are colored based on :attr:`Renderer.draw_color`).

        If configured correctly and supported by an underlying rendering driver, Renderer
        objects can have a :class:`Texture` object temporarily set as a target texture
        (the Texture object must have been created with target texture usage support),
        which allows those textures to be drawn onto.

        To present drawn content onto the window, :meth:`Renderer.present` should be
        called. :meth:`Renderer.clear` should be called to clear any drawn content
        with the set Renderer draw color.

        When things are drawn onto Renderer objects, an internal batching system is
        used by default to batch those "draw calls" together, to have all of them be
        processed in one go when :meth:`Renderer.present` is called. This is unlike
        :class:`pygame.Surface` objects, on which modifications via blitting occur
        immediately, but lends well to the behavior of GPUs, as draw calls can be
        expensive on lower-end models.
        """

        # https://wiki.libsdl.org/SDL_CreateRenderer
        # https://wiki.libsdl.org/SDL_RendererFlags
        flags = 0
        if accelerated >= 0:
            flags |= _SDL_RENDERER_ACCELERATED if accelerated else _SDL_RENDERER_SOFTWARE
        if vsync:
            flags |= _SDL_RENDERER_PRESENTVSYNC
        if target_texture:
            flags |= _SDL_RENDERER_TARGETTEXTURE

        self._renderer = SDL_CreateRenderer(window._win, index, flags)
        if not self._renderer:
            raise error()

        cdef Uint8[4] defaultColor = [255, 255, 255, 255]
        self._target = None
        self._win = window
        self._is_borrowed=0

    def __dealloc__(self):
        if self._is_borrowed:
            return
        if self._renderer:
            SDL_DestroyRenderer(self._renderer)

    @property
    def draw_blend_mode(self):
        """Get or set the blend mode used for primitive drawing operations
        """
        # https://wiki.libsdl.org/SDL_GetRenderDrawBlendMode
        cdef SDL_BlendMode blendMode
        cdef int res = SDL_GetRenderDrawBlendMode(self._renderer, &blendMode)
        if res < 0:
            raise error()

        return blendMode

    @draw_blend_mode.setter
    def draw_blend_mode(self, blendMode):
        # https://wiki.libsdl.org/SDL_SetRenderDrawBlendMode
        cdef int res = SDL_SetRenderDrawBlendMode(self._renderer, blendMode)
        if res < 0:
            raise error()

    @property
    def draw_color(self):
        """Get or set the color used for primitive drawing operations
        """
        cdef Uint8[4] rgba

        cdef int res = SDL_GetRenderDrawColor(self._renderer,
                                              &(rgba[0]),
                                              &(rgba[1]),
                                              &(rgba[2]),
                                              &(rgba[3]))

        if res < 0:
            raise error()

        return pgColor_NewLength(rgba, 4)

    @draw_color.setter
    def draw_color(self, new_value):
        cdef Uint8[4] rgba
        pg_RGBAFromObjEx(new_value, rgba, PG_COLOR_HANDLE_ALL)

        # https://wiki.libsdl.org/SDL_SetRenderDrawColor
        cdef int res = SDL_SetRenderDrawColor(self._renderer,
                                              rgba[0],
                                              rgba[1],
                                              rgba[2],
                                              rgba[3])

        if res < 0:
            raise error()

    def clear(self):
        """Clear the current rendering target with the drawing color
        """
        # https://wiki.libsdl.org/SDL_RenderClear
        cdef int res = SDL_RenderClear(self._renderer)
        if res < 0:
            raise error()

    def present(self):
        """Update the screen with any rendering performed since the previous call

        Presents the composed backbuffer to the screen.
        Updates the screen with any rendering performed since the previous call.
        """
        # https://wiki.libsdl.org/SDL_RenderPresent
        SDL_RenderPresent(self._renderer)

    cpdef get_viewport(self):
        """Get the drawing area on the rendering target
        """
        # https://wiki.libsdl.org/SDL_RenderGetViewport
        cdef SDL_Rect rect
        SDL_RenderGetViewport(self._renderer, &rect)
        return pgRect_New(&rect)

    def set_viewport(self, area):
        """Set the drawing area on the rendering target

        :param area: A :class:`pygame.Rect` or tuple representing the
                     drawing area on the target, or ``None`` to use the
                     entire area of the current rendering target.
        """
        # https://wiki.libsdl.org/SDL_RenderSetViewport
        if area is None:
            if SDL_RenderSetViewport(self._renderer, NULL) < 0:
                raise error()
            return

        cdef SDL_Rect tmprect
        cdef SDL_Rect *rectptr = pgRect_FromObject(area, &tmprect)
        if rectptr == NULL:
            raise TypeError('expected a rectangle')

        if SDL_RenderSetViewport(self._renderer, rectptr) < 0:
            raise error()

    @property
    def logical_size(self):
        """Get or set the logical Renderer size (a device independent resolution for rendering)
        """
        cdef int w
        cdef int h
        SDL_RenderGetLogicalSize(self._renderer, &w, &h)
        return (w, h)

    @logical_size.setter
    def logical_size(self, size):
        cdef int w = size[0]
        cdef int h = size[1]
        if (SDL_RenderSetLogicalSize(self._renderer, w, h) != 0):
            raise error()

    @property
    def scale(self):
        """Get the drawing scale for the current rendering target
        """
        cdef float x
        cdef float y
        SDL_RenderGetScale(self._renderer, &x, &y)
        return (x, y)

    @scale.setter
    def scale(self, scale):
        cdef float x = scale[0]
        cdef float y = scale[1]
        if (SDL_RenderSetScale(self._renderer, x, y) != 0):
            raise error()

    # TODO ifdef
    # def is_integer_scale(self):
    #     return SDL_RenderGetIntegerScale(self._renderer)

    @property
    def target(self):
        """`Get or set the current rendering target

        Gets or sets the current rendering target.
        A value of ``None`` means that no custom rendering target was set and the
        Renderer's window will be used as the target.
        """
        # https://wiki.libsdl.org/SDL_GetRenderTarget
        return self._target

    @target.setter
    def target(self, newtarget):
        # https://wiki.libsdl.org/SDL_SetRenderTarget
        if newtarget is None:
            self._target = None
            if SDL_SetRenderTarget(self._renderer, NULL) < 0:
                raise error()
        elif isinstance(newtarget, Texture):
            self._target = newtarget
            if SDL_SetRenderTarget(self._renderer,
                                   self._target._tex) < 0:
                raise error()
        else:
            raise TypeError('target must be a Texture or None')

    cpdef object blit(self, object source, Rect dest=None, Rect area=None, int special_flags=0):
        """Draw textures using a Surface-like API

        For compatibility purposes. Draws :class:`Texture` objects onto the
        Renderer using a method signature similar to :meth:`pygame.Surface.blit`.

        :param source: A :class:`Texture` or :class:`Image` to draw.
        :param dest: The drawing destination on the rendering target.
        :param area: The portion of the source texture or image to draw from.
        :param special_flags: have no effect at this moment.

        .. note:: Textures created by different Renderers cannot shared with each other!
        """
        if isinstance(source, Texture):
            (<Texture>source).draw(area, dest)
        elif isinstance(source, Image):
            (<Image>source).draw(area, dest)
        elif not hasattr(source, 'draw'):
            raise TypeError('source must be drawable')
        else:
            source.draw(area, dest)

        if not dest:
            return self.get_viewport()
        return dest

    def draw_line(self, p1, p2):
        """Draw a line

        :param p1: The line start point.
        :param p2: The line end point.
        """
        cdef int res
        res = SDL_RenderDrawLineF(self._renderer,
                                 p1[0], p1[1],
                                 p2[0], p2[1])
        if res < 0:
            raise error()

    def draw_point(self, point):
        """Draw a point

        :param point: The point's coordinate.
        """
        # https://wiki.libsdl.org/SDL_RenderDrawPointF
        cdef int res
        res = SDL_RenderDrawPointF(self._renderer,
                                point[0], point[1])
        if res < 0:
            raise error()

    def draw_rect(self, rect):
        """Draw a rectangle outline

        :param rect: The :class:`Rect`-like rectangle to draw.
        """
        # https://wiki.libsdl.org/SDL_RenderDrawRectF
        cdef SDL_FRect _frect
        cdef SDL_FRect *frectptr
        cdef int res

        frectptr = pgFRect_FromObject(rect, &_frect)
        if frectptr == NULL:
            raise TypeError('expected a rectangle')

        res = SDL_RenderDrawRectF(self._renderer, frectptr)

        if res < 0:
            raise error()

    def fill_rect(self, rect):
        """Draw a filled rectangle

        :param rect: The :class:`Rect`-like rectangle to draw.
        """
        # https://wiki.libsdl.org/SDL_RenderFillRectF
        cdef SDL_FRect _frect
        cdef SDL_FRect *frectptr
        cdef int res


        frectptr = pgFRect_FromObject(rect, &_frect)
        if frectptr == NULL:
            raise TypeError('expected a rectangle')

        res = SDL_RenderFillRectF(self._renderer, frectptr)

        if res < 0:
            raise error()

    def draw_triangle(self, p1, p2, p3):
        # https://wiki.libsdl.org/SDL_RenderDrawLinesF
        cdef SDL_FPoint fpoints[4]

        for i, pos in enumerate((p1, p2, p3, p1)):
            fpoints[i].x = pos[0]
            fpoints[i].y = pos[1]

        res = SDL_RenderDrawLinesF(self._renderer, fpoints, 4)

        if res < 0:
            raise error()

    def fill_triangle(self, p1, p2, p3):
        # https://wiki.libsdl.org/SDL_RenderGeometry
        if not SDL_VERSION_ATLEAST(2, 0, 18):
            raise error("fill_triangle requires SDL 2.0.18 or newer")

        cdef Uint8[4] rgba

        cdef int res = SDL_GetRenderDrawColor(self._renderer,
                                              &(rgba[0]),
                                              &(rgba[1]),
                                              &(rgba[2]),
                                              &(rgba[3]))

        if res < 0:
            raise error()

        cdef SDL_Vertex vertices[3]
        for i, pos in enumerate((p1, p2, p3)):
            vertices[i].position.x = pos[0]
            vertices[i].position.y = pos[1]
            vertices[i].color.r = rgba[0]
            vertices[i].color.g = rgba[1]
            vertices[i].color.b = rgba[2]
            vertices[i].color.a = rgba[3]

        res = SDL_RenderGeometry(self._renderer, NULL, vertices, 3, NULL, 0)
        if res < 0:
            raise error()

    def draw_quad(self, p1, p2, p3, p4):
        # https://wiki.libsdl.org/SDL_RenderDrawLinesF
        cdef SDL_FPoint fpoints[5]
        for i, pos in enumerate((p1, p2, p3, p4, p1)):
            fpoints[i].x = pos[0]
            fpoints[i].y = pos[1]

        res = SDL_RenderDrawLinesF(self._renderer, fpoints, 5)

        if res < 0:
            raise error()

    def fill_quad(self, p1, p2, p3, p4):
        # https://wiki.libsdl.org/SDL_RenderGeometry
        if not SDL_VERSION_ATLEAST(2, 0, 18):
            raise error("fill_quad requires SDL 2.0.18 or newer")

        cdef Uint8[4] rgba

        cdef int res = SDL_GetRenderDrawColor(self._renderer,
                                              &(rgba[0]),
                                              &(rgba[1]),
                                              &(rgba[2]),
                                              &(rgba[3]))

        if res < 0:
            raise error()

        cdef SDL_Vertex vertices[6]
        for i, pos in enumerate((p1, p2, p3, p3, p4, p1)):
            vertices[i].position.x = pos[0]
            vertices[i].position.y = pos[1]
            vertices[i].color.r = rgba[0]
            vertices[i].color.g = rgba[1]
            vertices[i].color.b = rgba[2]
            vertices[i].color.a = rgba[3]

        res = SDL_RenderGeometry(self._renderer, NULL, vertices, 6, NULL, 0)
        if res < 0:
            raise error()

    def to_surface(self, surface=None, area=None):
        # https://wiki.libsdl.org/SDL_RenderReadPixels
        """Read pixels from current rendering target and create a Surface (slow operation, use sparingly)

        Read pixel data from the current rendering target and return a
        :class:`pygame.Surface` containing it.

        :param Surface surface: A :class:`pygame.Surface` object to read the pixel
                                data into. It must be large enough to fit the area, otherwise
                                ``ValueError`` is raised.
                                If set to ``None``, a new surface will be created.
        :param area: The area of the screen to read pixels from. The area is
                     clipped to fit inside the viewport.
                     If ``None``, the entire viewport is used.

        .. note::
            This is a very slow operation, due to the overhead of the VRAM to RAM
            data transfer and the cost of creating a potentially large
            :class:`pygame.Surface`. It should not be used frequently.
        """
        cdef Uint32 format
        cdef SDL_Rect rarea
        cdef SDL_Rect tempviewport
        cdef SDL_Rect *areaparam
        cdef SDL_Surface *surf
        cdef SDL_Rect *rectptr

        # obtain area to use
        if area is not None:

            rectptr = pgRect_FromObject(area, &rarea)
            if rectptr == NULL:
                raise TypeError('area must be None or a rect')

            # clip area
            SDL_RenderGetViewport(self._renderer, &tempviewport)
            SDL_IntersectRect(rectptr, &tempviewport, rectptr)

            areaparam = rectptr
            rarea.x = rectptr.x
            rarea.y = rectptr.y
            rarea.w = rectptr.w
            rarea.h = rectptr.h
        else:
            SDL_RenderGetViewport(self._renderer, &rarea)
            areaparam = NULL

        # prepare surface
        if surface is None:
            # create a new surface
            format = SDL_GetWindowPixelFormat(self._win._win)
            if format == SDL_PIXELFORMAT_UNKNOWN:
                raise error()

            surf = SDL_CreateRGBSurfaceWithFormat(
                0,
                rarea.w, rarea.h,
                SDL_BITSPERPIXEL(format),
                format)
            if surf == NULL:
                raise MemoryError("not enough memory for the surface")

            surface = <object>pgSurface_New2(surf, 1)
            # casting to <object> makes cython assume reference counting of the object and it increments
            # it by one, however, pgSurface_New2 already returns a new reference to an object, so
            # we need to decrement that reference cython has added on top
            # see https://github.com/cython/cython/issues/2589#issuecomment-417604249 for additional context
            Py_DECREF(surface)
        elif pgSurface_Check(surface):
            surf = pgSurface_AsSurface(surface)
            if surf.w < rarea.w or surf.h < rarea.h:
                raise ValueError("the surface is too small")
            format = surf.format.format
        else:
            raise TypeError("'surface' must be a surface or None")

        if SDL_RenderReadPixels(self._renderer,
                                areaparam,
                                format, surf.pixels, surf.pitch) < 0:
            raise error()
        return surface

    @staticmethod
    def compose_custom_blend_mode(color_mode, alpha_mode):
        """Compose a custom blend mode

        Compose a custom blend mode.
        See https://wiki.libsdl.org/SDL2/SDL_ComposeCustomBlendMode for more information.

        :param color_mode: A tuple ``(srcColorFactor, dstColorFactor, colorOperation)``
        :param alpha_mode: A tuple ``(srcAlphaFactor, dstAlphaFactor, alphaOperation)``

        :return: A blend mode to be used with :meth:`Renderer.set_draw_blend_mode` and :meth:`Texture.set_blend_mode`.
        """
        # https://wiki.libsdl.org/SDL_ComposeCustomBlendMode
        cdef int res = SDL_ComposeCustomBlendMode(color_mode[0],
                                                  color_mode[1],
                                                  color_mode[2],
                                                  alpha_mode[0],
                                                  alpha_mode[1],
                                                  alpha_mode[2])
        if res < 0:
            raise error()
        return res



import pygame
from math import pi

# Initialize pygame
pygame.init()

# Set the height and width of the screen
size = [400, 300]
screen = pygame.display.set_mode(size)

pygame.display.set_caption("Example code for the draw module")

# Loop until the user clicks the close button.
done = False
clock = pygame.time.Clock()

while not done:
    # This limits the while loop to a max of 60 times per second.
    # Leave this out and we will use all CPU we can.
    clock.tick(60)

    for event in pygame.event.get():  # User did something
        if event.type == pygame.QUIT:  # If user clicked close
            done = True  # Flag that we are done so we exit this loop

    # Clear the screen and set the screen background
    screen.fill("white")

    # Draw on the screen a green line from (0, 0) to (50, 30)
    # 5 pixels wide. Uses (r, g, b) color - medium sea green.
    pygame.draw.line(screen, (60, 179, 113), [0, 0], [50, 30], 5)

    # Draw on the screen a green antialiased line from (0, 25) to (50, 55)
    # 5 pixels wide. Uses (r, g, b) color - medium sea green.
    pygame.draw.aaline(screen, (60, 179, 113), [0, 25], [50, 55], 5)

    # Draw on the screen a green line from (0, 50) to (50, 80)
    # Because it is an antialiased line, it is 1 pixel wide.
    # Uses (r, g, b) color - medium sea green.
    pygame.draw.aaline(screen, (60, 179, 113), [0, 50], [50, 80], True)

    # Draw on the screen 3 black lines, each 5 pixels wide.
    # The 'False' means the first and last points are not connected.
    pygame.draw.lines(
        screen, "black", False, [[0, 80], [50, 90], [200, 80], [220, 30]], 5
    )

    # Draw a rectangle outline
    pygame.draw.rect(screen, "black", [75, 10, 50, 20], 2)

    # Draw a solid rectangle. Same color as "black" above, specified in a new way
    pygame.draw.rect(screen, (0, 0, 0), [150, 10, 50, 20])

    # Draw a rectangle with rounded corners
    pygame.draw.rect(screen, "green", [115, 210, 70, 40], 10, border_radius=15)
    pygame.draw.rect(
        screen,
        "red",
        [135, 260, 50, 30],
        0,
        border_radius=10,
        border_top_left_radius=0,
        border_bottom_right_radius=15,
    )

    # Draw an ellipse outline, using a rectangle as the outside boundaries
    pygame.draw.ellipse(screen, "red", [225, 10, 50, 20], 2)

    # Draw an solid ellipse, using a rectangle as the outside boundaries
    pygame.draw.ellipse(screen, "red", [300, 10, 50, 20])

    # This draws a triangle using the polygon command
    pygame.draw.polygon(screen, "black", [[100, 100], [0, 200], [200, 200]], 5)

    # Draw an arc as part of an ellipse.
    # Use radians to determine what angle to draw.
    pygame.draw.arc(screen, "black", [210, 75, 150, 125], 0, pi / 2, 2)
    pygame.draw.arc(screen, "green", [210, 75, 150, 125], pi / 2, pi, 2)
    pygame.draw.arc(screen, "blue", [210, 75, 150, 125], pi, 3 * pi / 2, 2)
    pygame.draw.arc(screen, "red", [210, 75, 150, 125], 3 * pi / 2, 2 * pi, 2)

    # Draw a circle
    pygame.draw.circle(screen, "blue", [60, 250], 40)

    # Draw an antialiased circle with 3 pixels wide line
    pygame.draw.aacircle(screen, "green", [340, 250], 40, 3)

    # Draw an antialiased top right circle quadrant with 4 pixels wide line
    pygame.draw.aacircle(screen, "red", [340, 250], 20, 4, draw_top_right=True)

    # Draw an antialiased bottom left filled circle quadrant
    pygame.draw.aacircle(screen, "blue", [340, 250], 20, draw_bottom_left=True)

    # Draw only one circle quadrant
    pygame.draw.circle(screen, "blue", [250, 250], 40, 0, draw_top_right=True)
    pygame.draw.circle(screen, "red", [250, 250], 40, 30, draw_top_left=True)
    pygame.draw.circle(screen, "green", [250, 250], 40, 20, draw_bottom_left=True)
    pygame.draw.circle(screen, "black", [250, 250], 40, 10, draw_bottom_right=True)

    # Go ahead and update the screen with what we've drawn.
    # This MUST happen after all the other drawing commands.
    pygame.display.flip()

# Be IDLE friendly
pygame.quit()



#!/usr/bin/env python
"""pygame.examples.setmodescale

On high resolution displays(4k, 1080p) and tiny graphics games (640x480)
show up very small so that they are unplayable. SCALED scales up the window
for you. The game thinks it's a 640x480 window, but really it can be bigger.
Mouse events are scaled for you, so your game doesn't need to do it.

Passing SCALED to pygame.display.set_mode means the resolution depends
on desktop size and the graphics are scaled.
"""

import pygame
import sys

RES = (160, 120)
FPS = 30


def main():
    pygame.init()

    clock = pygame.Clock()

    print("desktops", pygame.display.get_desktop_sizes())

    do_vsync = bool("--vsync" in sys.argv)

    if do_vsync:
        screen = pygame.display.set_mode(RES, pygame.SCALED | pygame.RESIZABLE, vsync=1)
    else:
        screen = pygame.display.set_mode(RES, pygame.SCALED | pygame.RESIZABLE)

    # MAIN LOOP

    done = False

    i = 0
    j = 0

    r_name, r_flags = pygame.display._get_renderer_info()
    print("renderer:", r_name, "flags:", bin(r_flags))
    for flag, name in [
        (1, "software"),
        (2, "accelerated"),
        (4, "VSync"),
        (8, "render to texture"),
    ]:
        if flag & r_flags:
            print(name)

    while not done:
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN and event.key == pygame.K_q:
                done = True
            if event.type == pygame.QUIT:
                done = True
            if event.type == pygame.KEYDOWN and event.key == pygame.K_f:
                pygame.display.toggle_fullscreen()
        i += 1
        i = i % screen.get_width()
        j += i % 2
        j = j % screen.get_height()

        screen.fill((255, 0, 255))
        pygame.draw.circle(screen, (0, 0, 0), (100, 100), 20)
        pygame.draw.circle(screen, (0, 0, 200), (0, 0), 10)
        pygame.draw.circle(screen, (200, 0, 0), (160, 120), 30)
        if do_vsync:
            # vertical line that moves horizontally to make screen tearing obvious
            pygame.draw.line(screen, (250, 250, 0), (i, 0), (i, 120))
        else:
            pygame.draw.line(screen, (250, 250, 0), (0, 120), (160, 0))
        pygame.draw.circle(screen, (255, 255, 255), (i, j), 5)

        pygame.display.set_caption("FPS:" + str(clock.get_fps()))
        if do_vsync:
            pygame.display.flip()
            # FPS should be limited by vsync, so we tick really fast
            # we only need to have the clock tick to track FPS
            clock.tick()
        else:
            clock.tick(FPS)
            pygame.display.flip()

    pygame.quit()


if __name__ == "__main__":
    main()



import sys
from collections.abc import Callable, Collection, Iterator
from typing import (
    Literal,
    SupportsIndex,
    TypeVar,
    Union,
    overload,
    Optional,
)
from typing_extensions import deprecated  # added in 3.13

from pygame.typing import Point, RectLike, SequenceLike

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# 'ellipsis' existed in typeshed pre 3.10, now we use EllipsisType which is
# the modern standard library equivalent.
if sys.version_info >= (3, 10):
    from types import EllipsisType
else:
    EllipsisType = ellipsis

_N = TypeVar("_N", int, float)
_K = TypeVar("_K")
_V = TypeVar("_V")
_T = TypeVar("_T")

_RectTypeCompatible_co = TypeVar("_RectTypeCompatible_co", bound=RectLike, covariant=True)

class _GenericRect(Collection[_N]):
    @property
    def x(self) -> _N: ...
    @x.setter
    def x(self, value: float) -> None: ...
    @property
    def y(self) -> _N: ...
    @y.setter
    def y(self, value: float) -> None: ...
    @property
    def top(self) -> _N: ...
    @top.setter
    def top(self, value: float) -> None: ...
    @property
    def left(self) -> _N: ...
    @left.setter
    def left(self, value: float) -> None: ...
    @property
    def bottom(self) -> _N: ...
    @bottom.setter
    def bottom(self, value: float) -> None: ...
    @property
    def right(self) -> _N: ...
    @right.setter
    def right(self, value: float) -> None: ...
    @property
    def topleft(self) -> tuple[_N, _N]: ...
    @topleft.setter
    def topleft(self, value: Point) -> None: ...
    @property
    def bottomleft(self) -> tuple[_N, _N]: ...
    @bottomleft.setter
    def bottomleft(self, value: Point) -> None: ...
    @property
    def topright(self) -> tuple[_N, _N]: ...
    @topright.setter
    def topright(self, value: Point) -> None: ...
    @property
    def bottomright(self) -> tuple[_N, _N]: ...
    @bottomright.setter
    def bottomright(self, value: Point) -> None: ...
    @property
    def midtop(self) -> tuple[_N, _N]: ...
    @midtop.setter
    def midtop(self, value: Point) -> None: ...
    @property
    def midleft(self) -> tuple[_N, _N]: ...
    @midleft.setter
    def midleft(self, value: Point) -> None: ...
    @property
    def midbottom(self) -> tuple[_N, _N]: ...
    @midbottom.setter
    def midbottom(self, value: Point) -> None: ...
    @property
    def midright(self) -> tuple[_N, _N]: ...
    @midright.setter
    def midright(self, value: Point) -> None: ...
    @property
    def center(self) -> tuple[_N, _N]: ...
    @center.setter
    def center(self, value: Point) -> None: ...
    @property
    def centerx(self) -> _N: ...
    @centerx.setter
    def centerx(self, value: float) -> None: ...
    @property
    def centery(self) -> _N: ...
    @centery.setter
    def centery(self, value: float) -> None: ...
    @property
    def size(self) -> tuple[_N, _N]: ...
    @size.setter
    def size(self, value: Point) -> None: ...
    @property
    def width(self) -> _N: ...
    @width.setter
    def width(self, value: float) -> None: ...
    @property
    def height(self) -> _N: ...
    @height.setter
    def height(self, value: float) -> None: ...
    @property
    def w(self) -> _N: ...
    @w.setter
    def w(self, value: float) -> None: ...
    @property
    def h(self) -> _N: ...
    @h.setter
    def h(self, value: float) -> None: ...
    __hash__: None  # type: ignore
    __safe_for_unpickling__: Literal[True]
    @overload
    def __init__(
        self, left: float, top: float, width: float, height: float
    ) -> None: ...
    @overload
    def __init__(self, left_top: Point, width_height: Point) -> None: ...
    @overload
    def __init__(self, single_arg: RectLike) -> None: ...
    @overload
    def __init__(self) -> None: ...
    def __len__(self) -> Literal[4]: ...
    def __iter__(self) -> Iterator[_N]: ...
    @overload
    def __getitem__(self, i: SupportsIndex) -> _N: ...
    @overload
    def __getitem__(self, s: Union[slice, EllipsisType]) -> list[_N]: ...
    @overload
    def __setitem__(self, key: int, value: float) -> None: ...
    @overload
    def __setitem__(
        self, key: Union[slice, EllipsisType], value: Union[float, RectLike]
    ) -> None: ...
    def __copy__(self) -> Self: ...
    def copy(self) -> Self: ...
    @overload
    def move(self, x: float, y: float, /) -> Self: ...
    @overload
    def move(self, move_by: Point, /) -> Self: ...
    @overload
    def move_ip(self, x: float, y: float, /) -> None: ...
    @overload
    def move_ip(self, move_by: Point, /) -> None: ...
    def move_to(self, **kwargs: Union[float, Point]) -> Self: ...
    @overload
    def inflate(self, x: float, y: float, /) -> Self: ...
    @overload
    def inflate(self, inflate_by: Point, /) -> Self: ...
    @overload
    def inflate_ip(self, x: float, y: float, /) -> None: ...
    @overload
    def inflate_ip(self, inflate_by: Point, /) -> None: ...
    @overload
    def scale_by(self, x: float, y: float = ...) -> Self: ...
    @overload
    def scale_by(self, scale_by: Point) -> Self: ...
    @overload
    def scale_by_ip(self, x: float, y: float = ...) -> None: ...
    @overload
    def scale_by_ip(self, scale_by: Point) -> None: ...
    @overload
    def update(self, left: float, top: float, width: float, height: float, /) -> None: ...
    @overload
    def update(self, left_top: Point, width_height: Point, /) -> None: ...
    @overload
    def update(self, single_arg: RectLike, /) -> None: ...
    @overload
    def clamp(self, rect: RectLike, /) -> Self: ...
    @overload
    def clamp(self, left_top: Point, width_height: Point, /) -> Self: ...
    @overload
    def clamp(self, left: float, top: float, width: float, height: float, /) -> Self: ...
    @overload
    def clamp_ip(self, rect: RectLike, /) -> None: ...
    @overload
    def clamp_ip(self, left_top: Point, width_height: Point, /) -> None: ...
    @overload
    def clamp_ip(
        self, left: float, top: float, width: float, height: float, /
    ) -> None: ...
    @overload
    def clip(self, rect: RectLike, /) -> Self: ...
    @overload
    def clip(self, left_top: Point, width_height: Point, /) -> Self: ...
    @overload
    def clip(self, left: float, top: float, width: float, height: float, /) -> Self: ...
    @overload
    def clipline(
        self, x1: float, x2: float, x3: float, x4: float, /
    ) -> Union[tuple[tuple[_N, _N], tuple[_N, _N]], tuple[()]]: ...
    @overload
    def clipline(
        self, first_point: Point, second_point: Point, /
    ) -> Union[tuple[tuple[_N, _N], tuple[_N, _N]], tuple[()]]: ...
    @overload
    def clipline(
        self, rect_arg: RectLike, /
    ) -> Union[tuple[tuple[_N, _N], tuple[_N, _N]], tuple[()]]: ...
    @overload
    def union(self, rect: RectLike, /) -> Self: ...
    @overload
    def union(self, left_top: Point, width_height: Point, /) -> Self: ...
    @overload
    def union(self, left: float, top: float, width: float, height: float, /) -> Self: ...
    @overload
    def union_ip(self, rect: RectLike, /) -> None: ...
    @overload
    def union_ip(self, left_top: Point, width_height: Point, /) -> None: ...
    @overload
    def union_ip(
        self, left: float, top: float, width: float, height: float, /
    ) -> None: ...
    def unionall(self, rect: SequenceLike[_RectTypeCompatible_co], /) -> Self: ...
    def unionall_ip(self, rect_SequenceLike: SequenceLike[_RectTypeCompatible_co], /) -> None: ...
    @overload
    def fit(self, rect: RectLike, /) -> Self: ...
    @overload
    def fit(self, left_top: Point, width_height: Point, /) -> Self: ...
    @overload
    def fit(self, left: float, top: float, width: float, height: float, /) -> Self: ...
    def normalize(self) -> None: ...
    def __contains__(self, rect: Union[RectLike, _N], /) -> bool: ...  # type: ignore[override]
    @overload
    def contains(self, rect: RectLike, /) -> bool: ...
    @overload
    def contains(self, left_top: Point, width_height: Point, /) -> bool: ...
    @overload
    def contains(
        self, left: float, top: float, width: float, height: float, /
    ) -> bool: ...
    @overload
    def collidepoint(self, x: float, y: float, /) -> bool: ...
    @overload
    def collidepoint(self, x_y: Point, /) -> bool: ...
    @overload
    def colliderect(self, rect: RectLike, /) -> bool: ...
    @overload
    def colliderect(self, left_top: Point, width_height: Point, /) -> bool: ...
    @overload
    def colliderect(
        self, left: float, top: float, width: float, height: float, /
    ) -> bool: ...
    def collidelist(self, rect_list: SequenceLike[_RectTypeCompatible_co], /) -> int: ...
    def collidelistall(self, rect_list: SequenceLike[_RectTypeCompatible_co], /) -> list[int]: ...
    def collideobjectsall(
        self, objects: SequenceLike[_T], key: Optional[Callable[[_T], RectLike]] = None
    ) -> list[_T]: ...
    def collideobjects(
        self, objects: SequenceLike[_T], key: Optional[Callable[[_T], RectLike]] = None
    ) -> Optional[_T]: ...
    @overload
    def collidedict(
        self, rect_dict: dict[_RectTypeCompatible_co, _V], values: Literal[False] = False
    ) -> Optional[tuple[_RectTypeCompatible_co, _V]]: ...
    @overload
    def collidedict(
        self, rect_dict: dict[_K, _RectTypeCompatible_co], values: Literal[True]
    ) -> Optional[tuple[_K, _RectTypeCompatible_co]]: ...
    @overload
    def collidedictall(
        self, rect_dict: dict[_RectTypeCompatible_co, _V], values: Literal[False] = False
    ) -> list[tuple[_RectTypeCompatible_co, _V]]: ...
    @overload
    def collidedictall(
        self, rect_dict: dict[_K, _RectTypeCompatible_co], values: Literal[True]
    ) -> list[tuple[_K, _RectTypeCompatible_co]]: ...

# Rect confirms to the Collection ABC, since it also confirms to
# Sized, Iterable and Container ABCs
class Rect(_GenericRect[int]):
    ...

class FRect(_GenericRect[float]):
    ...

@deprecated("Use `Rect` instead (RectType is an old alias)")
class RectType(Rect): ...

@deprecated("Use `FRect` instead (FRectType is an old alias)")
class FRectType(FRect): ...



"""
buildconfig/stubs/gen_stubs.py
A script to auto-generate locals.pyi, constants.pyi and __init__.pyi typestubs
"""

import pathlib
import shutil
from typing import Any

import pygame.constants
import pygame.locals

doc_as_comment = __doc__.replace("\n", "\n# ").strip() if __doc__ else ""
info_header = f"{doc_as_comment} IMPORTANT NOTE: Do not edit this file by hand!\n"

# pygame-ce submodules that are auto-imported must be kept in this list
# keep in mind that not all pygame-ce modules are auto-imported, hence not all
# pygame-ce submodules make it to this list
PG_AUTOIMPORT_SUBMODS = [
    "display",
    "draw",
    "event",
    "font",
    "image",
    "key",
    "mixer",
    "mouse",
    "time",
    "cursors",
    "joystick",
    "math",
    "mask",
    "pixelcopy",
    "sndarray",
    "sprite",
    "surfarray",
    "transform",
    "scrap",
    "version",
    "base",
    "bufferproxy",
    "color",
    "colordict",
    "mixer_music",
    "pixelarray",
    "rect",
    "rwobject",
    "surface",
    "surflock",
    "sysfont",
    "_debug",
    "system",
    "geometry",
    "window",
    "typing",
]

# pygame classes that are autoimported into main namespace are kept in this dict
PG_AUTOIMPORT_CLASSES = {
    "rect": ["Rect", "FRect"],
    "surface": ["Surface", "SurfaceType"],
    "color": ["Color"],
    "pixelarray": ["PixelArray"],
    "math": ["Vector2", "Vector3"],
    "cursors": ["Cursor"],
    "bufferproxy": ["BufferProxy"],
    "mask": ["Mask"],
    "_debug": ["print_debug_info"],
    "event": ["Event"],
    "font": ["Font"],
    "mixer": ["Sound", "Channel"],
    "time": ["Clock"],
    "joystick": ["Joystick"],
    "window": ["Window"],
    "base": ["__version__"],  # need an explicit import
    # uncomment below line if Circle is added to the base namespace later
    # "geometry": ["Circle"],
}

# pygame modules from which __init__.py does the equivalent of
# from submod import *
# should be kept here
PG_STAR_IMPORTS = ("base", "rwobject", "version", "constants")


def get_all(mod: Any):
    """
    Get the attributes that are imported from 'mod' when 'from mod import *'
    First try to use '__all__' if it is defined, else fallback to 'dir'
    """
    if hasattr(mod, "__all__") and isinstance(mod.__all__, list):
        return sorted({str(i) for i in mod.__all__})

    return [i for i in dir(mod) if not i.startswith("_")]


# store all imports of __init__.pyi
pygame_all_imports = {"pygame": PG_AUTOIMPORT_SUBMODS}
for k, v in PG_AUTOIMPORT_CLASSES.items():
    pygame_all_imports[f".{k}"] = v

for k in PG_STAR_IMPORTS:
    val = get_all(getattr(pygame, k))
    try:
        pygame_all_imports[f".{k}"].extend(val)
    except KeyError:
        pygame_all_imports[f".{k}"] = val

# misc stubs that must be added to __init__.pyi
misc_stubs = """"""

# write constants.pyi file
constants_file = pathlib.Path(__file__).parent / "pygame" / "constants.pyi"
with open(constants_file, "w") as f:
    # write the module docstring of this file in the generated file, so that
    # people know this file exists
    f.write(info_header)
    f.write("\n")

    for element in pygame_all_imports[".constants"]:
        constant_type = getattr(pygame.constants, element).__class__.__name__
        f.write(f"{element}: {constant_type}\n")


# write __init__.pyi file
init_file = pathlib.Path(__file__).parent / "pygame" / "__init__.pyi"
with open(init_file, "w") as f:
    # write the module docstring of this file in the generated file, so that
    # people know this file exists
    f.write(info_header)
    f.write(misc_stubs)

    for mod, items in pygame_all_imports.items():
        if mod == "pygame":
            mod = "."
        if len(items) <= 4:
            # try to write imports in a single line if it can fit the line limit
            import_items = (f"{string} as {string}" for string in items)
            import_line = f"\nfrom {mod} import {', '.join(import_items)}"
            if len(import_line) <= 88:
                f.write(import_line)
                continue

        f.write(f"\nfrom {mod} import (\n")
        for item in items:
            f.write(f"    {item} as {item},\n")
        f.write(")\n")

# write locals.pyi file
locals_file = pathlib.Path(__file__).parent / "pygame" / "locals.pyi"
with open(locals_file, "w") as f:
    f.write(info_header)
    f.write("\n")

    for element in get_all(pygame.locals):
        constant_type = getattr(pygame.locals, element).__class__.__name__
        f.write(f"{element}: {constant_type}\n")

# copy typing.py to typing.pyi for type checkers
typing_py_file = pathlib.Path(__file__).parent.parent.parent / "src_py" / "typing.py"
typing_stub_file = pathlib.Path(__file__).parent / "pygame" / "typing.pyi"
shutil.copyfile(typing_py_file, typing_stub_file)



#!/usr/bin/env python
"""pygame.examples.ninepatch
Demonstrate the purpose of the 9-patch scale method and a way to implement it.

9-patch scaling allows us to scale a surface while keeping 4 corners intact.
This is very useful for GUI applications.

How it works:
The 4 corners are extracted from the original surface and pasted on the new surface.
The 4 sides in between the corners are scaled horizontally or vertically to fit
between the new surface corners.
The central part of the image is scaled and pasted.

In this example a green fill is used which can be done faster and easier with pygame.draw.rect
but usually you'd have a custom surface that needs intact corners.
"""

import os
import pygame

SCREEN_SIZE = pygame.Vector2(600, 500)
SCALE_SIZE = pygame.Vector2(500, 150)
INNER_SCALE_SIZE = pygame.Vector2(400, 115)
CORNER_SIZE = 20
INNER_CORNER_SIZE = 7


def ninepatch_scale(
    surface: pygame.Surface,
    size: pygame.typing.Point,
    corner_size: int,
    alpha: bool = True,
    smooth: bool = False,
    dest_surface: pygame.Surface | None = None,
) -> pygame.Surface:
    # the code of this function also emulates how a transform-like ninepatch function would likely behave
    if dest_surface is not None:
        ret_surface = dest_surface
        if ret_surface.size != size:
            raise ValueError("Destination surface doesn't match the provided size")
    else:
        ret_surface = pygame.Surface(
            size, pygame.SRCALPHA * alpha
        )  # when alpha is False the flags become 0

    # aliases
    ret_w, ret_h = size  # non-sequence argument caught by python
    src_w, src_h = surface.size
    c = corner_size
    scale_func = pygame.transform.smoothscale if smooth else pygame.transform.scale
    subsurface = surface.subsurface  # alias for performance/clarity

    if corner_size < 0 or corner_size > min(ret_w, ret_h):
        raise ValueError(
            "Corner size must be nonnegative and not greater than the smaller between width and height"
        )
    if corner_size == 0:  # default to normal scaling
        return scale_func(surface, size)

    src_x_side, src_y_side = surface.size - pygame.Vector2(c) * 2
    ret_x_side, ret_y_side = size - pygame.Vector2(c) * 2

    ret_surface.blit(  # topleft corner
        subsurface(0, 0, c, c), (0, 0)
    )
    ret_surface.blit(  # topright corner
        subsurface(src_w - c, 0, c, c), (ret_w - c, 0)
    )
    ret_surface.blit(  # bottomleft corner
        subsurface(0, src_h - c, c, c), (0, ret_h - c)
    )
    ret_surface.blit(  # bottomright corner
        subsurface(src_w - c, src_h - c, c, c), (ret_w - c, ret_h - c)
    )

    if src_x_side > 0:
        ret_surface.blit(  # top side
            scale_func(subsurface(c, 0, src_x_side, c), (ret_x_side, c)), (c, 0)
        )
        ret_surface.blit(  # bottom side
            scale_func(subsurface(c, src_h - c, src_x_side, c), (ret_x_side, c)),
            (c, ret_h - c),
        )

    if src_y_side > 0:
        ret_surface.blit(  # left side
            scale_func(subsurface(0, c, c, src_y_side), (c, ret_y_side)), (0, c)
        )
        ret_surface.blit(  # right side
            scale_func(subsurface(src_w - c, c, c, src_y_side), (c, ret_y_side)),
            (ret_w - c, c),
        )

    if src_x_side > 0 and src_y_side > 0:
        ret_surface.blit(  # central area
            scale_func(
                subsurface(c, c, src_x_side, src_y_side), (ret_x_side, ret_y_side)
            ),
            (c, c),
        )

    return ret_surface


def main():
    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE)
    pygame.display.set_caption("9-Patch Scale and Normal Scale Example")
    clock = pygame.Clock()
    font = pygame.Font(None, 30)

    main_dir = os.path.split(os.path.abspath(__file__))[0]
    example_image = pygame.image.load(os.path.join(main_dir, "data", "frame.png"))

    original_surface = pygame.Surface((100, 100), pygame.SRCALPHA)
    pygame.draw.rect(
        original_surface, "green", original_surface.get_rect(), 0, CORNER_SIZE
    )

    normal_scale = pygame.transform.scale(original_surface, SCALE_SIZE)
    normal_center = SCREEN_SIZE.elementwise() / pygame.Vector2(2, 4)
    normal_rect = normal_scale.get_rect(center=normal_center)
    example_normal_scale = pygame.transform.scale(example_image, INNER_SCALE_SIZE)
    example_normal_rect = example_normal_scale.get_rect(center=normal_center)

    ninepatch = ninepatch_scale(
        original_surface, SCALE_SIZE, CORNER_SIZE, True, False, None
    )
    ninepatch_center = SCREEN_SIZE.elementwise() / pygame.Vector2(2, 4 / 3)
    ninepatch_rect = ninepatch.get_rect(center=ninepatch_center)
    example_ninepatch = ninepatch_scale(
        example_image, INNER_SCALE_SIZE, INNER_CORNER_SIZE, True, False, None
    )
    example_ninepatch_rect = example_ninepatch.get_rect(center=ninepatch_center)

    text_normal = font.render("Normal Scale", True, "black")
    text_normal_rect = text_normal.get_rect(center=normal_center)
    text_9patch = font.render("9-Patch Scale", True, "black")
    text_9patch_rect = text_9patch.get_rect(center=ninepatch_center)

    running = True
    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False

        screen.fill(0)

        screen.blit(normal_scale, normal_rect)
        screen.blit(example_normal_scale, example_normal_rect)
        screen.blit(ninepatch, ninepatch_rect)
        screen.blit(example_ninepatch, example_ninepatch_rect)
        screen.blit(text_normal, text_normal_rect)
        screen.blit(text_9patch, text_9patch_rect)

        pygame.display.flip()
        clock.tick(60)

    pygame.quit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""pygame.examples.music_drop_fade
Fade in and play music from a list while observing several events

Adds music files to a playlist whenever played by one of the following methods
Music files passed from the commandline are played
Music files and filenames are played when drag and dropped onto the pygame window
Polls the clipboard and plays music files if it finds one there

Keyboard Controls:
* Press space or enter to pause music playback
* Press up or down to change the music volume
* Press left or right to seek 5 seconds into the track
* Press escape to quit
* Press any other button to skip to the next music file in the list
"""

import pygame
import os
import sys

VOLUME_CHANGE_AMOUNT = 0.02  # how fast should up and down arrows change the volume?
SCREEN_SIZE = (640, 480)

MUSIC_DONE = pygame.event.custom_type()  # event to be set as mixer.music.set_endevent()
main_dir = os.path.split(os.path.abspath(__file__))[0]
data_dir = os.path.join(main_dir, "data")

starting_pos = 0  # needed to fast forward and rewind
volume = 0.75
music_file_list = []
music_file_types = ("mp3", "ogg", "mid", "mod", "it", "xm", "wav")
music_can_seek = ("mp3", "ogg", "mod", "it", "xm")
screen = None


def add_file(filename):
    """
    This function will check if filename exists and is a music file
    If it is the file will be added to a list of music files(even if already there)
    Type checking is by the extension of the file, not by its contents
    We can only discover if the file is valid when we mixer.music.load() it later

    It looks in the file directory and its data subdirectory
    """
    if filename.rpartition(".")[2].lower() not in music_file_types:
        print(f"{filename} not added to file list")
        print("only these files types are allowed: ", music_file_types)
        return False
    elif os.path.exists(filename):
        music_file_list.append(filename)
    elif os.path.exists(os.path.join(main_dir, filename)):
        music_file_list.append(os.path.join(main_dir, filename))
    elif os.path.exists(os.path.join(data_dir, filename)):
        music_file_list.append(os.path.join(data_dir, filename))
    else:
        print("file not found")
        return False
    print(f"{filename} added to file list")
    return True


def play_file(filename):
    """
    This function will call add_file and play it if successful
    The music will fade in during the first 4 seconds
    set_endevent is used to post a MUSIC_DONE event when the song finishes
    The main loop will call play_next() when the MUSIC_DONE event is received
    """
    global starting_pos

    if add_file(filename):
        try:  # we must do this in case the file is not a valid audio file
            pygame.mixer.music.load(music_file_list[-1])
        except pygame.error as e:
            print(e)  # print description such as 'Not an Ogg Vorbis audio stream'
            if filename in music_file_list:
                music_file_list.remove(filename)
                print(f"{filename} removed from file list")
            return
        pygame.mixer.music.play(fade_ms=4000)
        pygame.mixer.music.set_volume(volume)

        if filename.rpartition(".")[2].lower() in music_can_seek:
            print("file supports seeking")
            starting_pos = 0
        else:
            print("file does not support seeking")
            starting_pos = -1
        pygame.mixer.music.set_endevent(MUSIC_DONE)


def play_next():
    """
    This function will play the next song in music_file_list
    It uses pop(0) to get the next song and then appends it to the end of the list
    The song will fade in during the first 4 seconds
    """

    global starting_pos
    if len(music_file_list) > 1:
        nxt = music_file_list.pop(0)

        try:
            pygame.mixer.music.load(nxt)
        except pygame.error as e:
            print(e)
            print(f"{nxt} removed from file list")

        music_file_list.append(nxt)
        print("starting next song: ", nxt)
    else:
        nxt = music_file_list[0]
    pygame.mixer.music.play(fade_ms=4000)
    pygame.mixer.music.set_volume(volume)
    pygame.mixer.music.set_endevent(MUSIC_DONE)

    if nxt.rpartition(".")[2].lower() in music_can_seek:
        starting_pos = 0
    else:
        starting_pos = -1


def change_music_position(amount):
    """
    Changes current playback position by amount seconds.
    This only works with OGG and MP3 files.
    music.get_pos() returns how many milliseconds the song has played, not
    the current position in the file. We must track the starting position
    ourselves. music.set_pos() will set the position in seconds.
    """
    global starting_pos

    if starting_pos >= 0:  # will be -1 unless play_file() was OGG or MP3
        played_for = pygame.mixer.music.get_pos() / 1000.0
        old_pos = starting_pos + played_for
        starting_pos = old_pos + amount
        pygame.mixer.music.play(start=starting_pos)
        print(f"jumped from {old_pos} to {starting_pos}")


def main():
    global font  # this will be used by the draw_text_line function
    global volume, starting_pos
    global screen
    running = True
    paused = False

    # we will be polling for key up and key down events
    # users should be able to change the volume by holding the up and down arrows
    # the change_volume variable will be set by key down events and cleared by key up events
    change_volume = 0

    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE)
    font = pygame.font.SysFont("Arial", 24)
    font.align = pygame.FONT_CENTER
    clock = pygame.Clock()

    clipped = ""
    if pygame.scrap.has_text():
        clipped = pygame.scrap.get_text()

    # add the command line arguments to the  music_file_list
    for arg in sys.argv[1:]:
        add_file(arg)
    play_file("house_lo.ogg")  # play default music included with pygame

    # draw instructions on screen
    text = """
Drop music files or path names onto this window
Copy file names into the clipboard
Or feed them from the command line
If it's music it will play!

LEFT and RIGHT will skip around the track
UP and DOWN will change volume
ENTER or SPACE will pause/unpause the track
Other keys will start the next track"""
    screen.blit(font.render(text, True, "white", "black", SCREEN_SIZE[0] - 20), (10, 5))
    """
    This is the main loop
    It will respond to drag and drop, clipboard changes, and key presses
    """
    while running:
        for ev in pygame.event.get():
            if ev.type == pygame.QUIT:
                running = False
            elif ev.type == pygame.DROPTEXT:
                play_file(ev.text)
            elif ev.type == pygame.DROPFILE:
                play_file(ev.file)
            elif ev.type == MUSIC_DONE:
                play_next()
            elif ev.type == pygame.KEYDOWN:
                if ev.key == pygame.K_ESCAPE:
                    running = False  # exit loop
                elif ev.key in (pygame.K_SPACE, pygame.K_RETURN):
                    if paused:
                        pygame.mixer.music.unpause()
                        paused = False
                    else:
                        pygame.mixer.music.pause()
                        paused = True
                elif ev.key == pygame.K_UP:
                    change_volume = VOLUME_CHANGE_AMOUNT
                elif ev.key == pygame.K_DOWN:
                    change_volume = -VOLUME_CHANGE_AMOUNT
                elif ev.key == pygame.K_RIGHT:
                    change_music_position(+5)
                elif ev.key == pygame.K_LEFT:
                    change_music_position(-5)

                else:
                    play_next()

            elif ev.type == pygame.KEYUP:
                if ev.key in (pygame.K_UP, pygame.K_DOWN):
                    change_volume = 0

        # is the user holding up or down?
        if change_volume:
            volume += change_volume
            volume = min(max(0, volume), 1)  # volume should be between 0 and 1
            pygame.mixer.music.set_volume(volume)
            print("volume:", volume)

        if pygame.scrap.has_text():
            new_text = pygame.scrap.get_text()
            if new_text != clipped:
                clipped = new_text
                play_file(clipped)

        pygame.display.flip()
        clock.tick(9)  # keep CPU use down by updating screen less often

    pygame.quit()


if __name__ == "__main__":
    main()



if __import__("sys").platform not in ("wasi", "emscripten"):
    from .sdl2 import *  # pylint: disable=wildcard-import; lgtm[py/polluting-import]
    from .audio import *  # pylint: disable=wildcard-import; lgtm[py/polluting-import]
    from .video import *  # pylint: disable=wildcard-import; lgtm[py/polluting-import]



import unittest, unittest.mock
import io

import pygame


class DebugTest(unittest.TestCase):
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def assert_stdout(self, expected_output, mock_stdout):
        pygame.print_debug_info()
        self.assertEqual(mock_stdout.getvalue(), expected_output)

    def test_print_debug(self):
        import os

        pygame.print_debug_info("temp_file.txt")
        with open("temp_file.txt", "r") as temp_file:
            text = temp_file.read()

        self.assertNotEqual(text, "")
        self.assert_stdout(text + "\n")

        os.remove("temp_file.txt")



from typing import Optional
from typing_extensions import deprecated # added in 3.13
from collections.abc import ByteString

@deprecated("since 2.2.0. Use the new API instead, which only requires display init")
def init() -> None: ...
@deprecated("since 2.2.0. Use the new API instead, which doesn't require scrap init")
def get_init() -> bool: ...
@deprecated("since 2.2.0. Use the new API instead: `pygame.scrap.get_text`")
def get(data_type: str, /) -> Optional[bytes]: ...
@deprecated("since 2.2.0. Use the new API instead, which only supports strings")
def get_types() -> list[str]: ...
@deprecated("since 2.2.0. Use the new API instead: `pygame.scrap.put_text`")
def put(data_type: str, data: ByteString, /) -> None: ...
@deprecated("since 2.2.0. Use the new API instead: `pygame.scrap.has_text`")
def contains(data_type: str, /) -> bool: ...
@deprecated("since 2.2.0. Use the new API instead, which uses system clipboard")
def lost() -> bool: ...
@deprecated("since 2.2.0. Use the new API instead, which only supports strings")
def set_mode(mode: int, /) -> None: ...
def put_text(text: str, /) -> None: ...
def get_text() -> str: ...
def has_text() -> bool: ...



#!/usr/bin/env python
"""pygame.examples.playmus

A simple music player.

   Use pygame.mixer.music to play an audio file.

A window is created to handle keyboard events for playback commands.


Keyboard Controls
-----------------

space - play/pause toggle
r     - rewind
f     - fade out
q     - stop

"""

import sys

import pygame

pygame.init()


def write_lines(
    dest: pygame.Surface,
    text: str,
    font=pygame.Font(None, 30),
    color=(254, 231, 21, 255),
    starting_line=0,
):
    text_surface = font.render(text, True, color)

    y_offset = 0
    if starting_line >= 0:
        y_offset = font.get_height() * starting_line
    else:
        y_offset = dest.get_height() + starting_line * font.get_height()

    dest.fill(
        (16, 24, 32, 255),
        pygame.Rect((0, y_offset), (text_surface.get_width(), dest.get_width())),
    )
    dest.blit(text_surface, (16, y_offset))


def show_usage_message():
    print(
        "Usage: python playmus.py <file>\n"
        "       python -m pygame.examples.playmus <file>"
    )


def main(file_path):
    """Play an audio file with pygame.mixer.music"""

    screen = pygame.display.set_mode((600, 400))
    screen.fill((16, 24, 32, 255))

    write_lines(screen, "Loading ...", starting_line=-1)
    pygame.mixer.init(frequency=44100)
    pygame.mixer.music.load(file_path)

    pygame.display.set_caption(f"File : {file_path}")

    write_lines(
        screen,
        "Press 'q' or 'ESCAPE' or close this window to quit\n"
        "Press 'SPACE' to play / pause\n"
        "Press 'r' to rewind to the beginning (restart)\n"
        "Press 'f' to fade music out over 5 seconds",
        starting_line=1,
    )

    pygame.mixer.music.play()
    write_lines(screen, "Playing ...\n", starting_line=-1)

    running = True
    paused = False

    while running:
        pygame.display.flip()

        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                key = event.key
                if key == pygame.K_SPACE:
                    if paused:
                        pygame.mixer.music.unpause()
                        paused = False
                        write_lines(screen, "Playing ...", starting_line=-1)
                    else:
                        pygame.mixer.music.pause()
                        paused = True
                        write_lines(screen, "Paused ...", starting_line=-1)
                elif key == pygame.K_r:
                    if file_path[-3:].lower() in ("ogg", "mp3", "mod"):
                        status = "Rewound."
                        pygame.mixer.music.rewind()
                    else:
                        status = "Restarted."
                        pygame.mixer.music.play()
                    if paused:
                        pygame.mixer.music.pause()
                        write_lines(screen, status, starting_line=-1)
                elif key == pygame.K_f:
                    write_lines(screen, "Fading out ...", starting_line=-1)
                    pygame.mixer.music.fadeout(5000)
                    # when finished get_busy() will return False.
                elif key in [pygame.K_q, pygame.K_ESCAPE]:
                    paused = False
                    pygame.mixer.music.stop()
                    # get_busy() will now return False.
            elif event.type == pygame.QUIT:
                running = False

        if not pygame.mixer.music.get_busy() and not paused:
            running = False

    pygame.quit()


if __name__ == "__main__":
    # Check the only command line argument, a file path
    if len(sys.argv) != 2:
        show_usage_message()
    else:
        main(sys.argv[1])



"""Pygame module for monitoring time.

Provides utilities for monitoring time, delaying time, and maintaining a
constant frame rate.
Times in pygame-ce are represented in milliseconds (1/1000 of a second).
"""

from typing import Union, final

from pygame.event import Event

def get_ticks() -> int:
    """Get the time in milliseconds.

    Return the number of milliseconds since ``pygame.init()`` was called. Before
    pygame is initialized this will always be 0.
    """

def wait(milliseconds: int, /) -> int:
    """Pause the program for an amount of time.

    Will pause for a given number of milliseconds. This function sleeps the
    process to share the processor with other programs. A program that waits for
    even a few milliseconds will consume very little processor time. It is
    slightly less accurate than the ``pygame.time.delay()`` function.

    This returns the actual number of milliseconds used.
    """

def delay(milliseconds: int, /) -> int:
    """Pause the program for an amount of time.

    Will pause for a given number of milliseconds. This function will use the
    processor (rather than sleeping) in order to make the delay more accurate
    than ``pygame.time.wait()``.

    This returns the actual number of milliseconds used.
    """

def set_timer(event: Union[int, Event], millis: int, loops: int = 0) -> None:
    """Repeatedly create an event on the event queue.

    Set an event to appear on the event queue every given number of milliseconds.
    The first event will not appear until the amount of time has passed.

    The ``event`` attribute can be a ``pygame.event.Event`` object or an integer
    type that denotes an event.

    ``loops`` is an integer that denotes the number of events posted. If 0 (default)
    then the events will keep getting posted, unless explicitly stopped.

    To disable the timer for such an event, call the function again with the same
    event argument with ``millis`` argument set to 0.

    It is also worth mentioning that a particular event type can only be put on a
    timer once. In other words, there cannot be two timers for the same event type.
    Setting an event timer for a particular event discards the old one for that
    event type.

    When this function is called with an ``Event`` object, the event(s) received
    on the event queue will be a shallow copy; the dict attribute of the event
    object passed as an argument and the dict attributes of the event objects
    received on timer will be references to the same dict object in memory.
    Modifications on one dict can affect another, use deepcopy operations on the
    dict object if you don't want this behaviour.
    However, calling this function with an integer event type would place event objects
    on the queue that don't have a common dict reference.

    ``loops`` replaces the ``once`` argument, and this does not break backward
    compatibility.

    .. versionaddedold:: 2.0.0.dev3 once argument added.
    .. versionchangedold:: 2.0.1 event argument supports ``pygame.event.Event`` object
    .. versionaddedold:: 2.0.1 added loops argument to replace once argument
    """

@final
class Clock:
    """Create an object to help track time.

    Creates a new Clock object that can be used to track an amount of time. The
    clock also provides several functions to help control a game's framerate.

    .. versionchanged:: 2.1.4  This class is also available through the ``pygame.Clock``
       alias.
    """

    def __new__(cls) -> Clock: ...
    def tick(self, framerate: float = 0, /) -> int:
        """Update the clock.

        This method should be called once per frame. It will compute how many
        milliseconds have passed since the previous call.

        If you pass the optional framerate argument the function will delay to
        keep the game running slower than the given ticks per second. This can be
        used to help limit the runtime speed of a game. By calling
        ``Clock.tick(40)`` once per frame, the program will never run at more
        than 40 frames per second.

        Note that this function uses SDL_Delay function which is not accurate on
        every platform, but does not use much CPU. Use tick_busy_loop if you want
        an accurate timer, and don't mind chewing CPU.
        """

    def tick_busy_loop(self, framerate: float = 0, /) -> int:
        """Update the clock.

        This method should be called once per frame. It will compute how many
        milliseconds have passed since the previous call.

        If you pass the optional framerate argument the function will delay to
        keep the game running slower than the given ticks per second. This can be
        used to help limit the runtime speed of a game. By calling
        ``Clock.tick_busy_loop(40)`` once per frame, the program will never run at
        more than 40 frames per second.

        Note that this function uses :func:`pygame.time.delay`, which uses lots
        of CPU in a busy loop to make sure that timing is more accurate.

        .. versionaddedold:: 1.8
        """

    def get_time(self) -> int:
        """Time used in the previous tick.

        The number of milliseconds that passed between the previous two calls to
        ``Clock.tick()``.
        """

    def get_rawtime(self) -> int:
        """Actual time used in the previous tick.

        Similar to ``Clock.get_time()``, but does not include any time used
        while ``Clock.tick()`` was delaying to limit the framerate.
        """

    def get_fps(self) -> float:
        """Compute the clock framerate.

        Compute your game's framerate (in frames per second). It is computed by
        averaging the last ten calls to ``Clock.tick()``.
        """



import sys
import pygame

WIDTH, HEIGHT = 500, 500


def connect_joystick(index, active_players, players, colors):
    if len(active_players) < len(players):
        joy = pygame.Joystick(index)
        index = players.index(None)
        # player data
        players[index] = {
            "joy": joy,
            "pos": [WIDTH * 0.25 + index * 64, 128],
            "surf_idx": index,
            "surf": colors[index],
            "joined": False,
        }
        active_players[joy.get_instance_id()] = index
        print(f"P{index + 1} Connected")


def disconnect_joystick(instance_id: int, active_players, players):
    index = active_players[instance_id]
    players[index] = None
    del active_players[instance_id]
    print(f"P{index + 1} Disconnected")


def control_player(player):  # move player
    joy = player["joy"]
    if not player["joined"]:
        return
    player["pos"][0] += joy.get_axis(0) * 5
    player["pos"][1] += joy.get_axis(1) * 5


def create_surf(size, color):
    surf = pygame.Surface(size)
    surf.fill(color)
    return surf


def main():
    pygame.init()

    screen = pygame.display.set_mode((WIDTH, HEIGHT))
    pygame.display.set_caption("Multiplayer Joystick example")
    clock = pygame.Clock()
    font_b = pygame.font.SysFont(None, 25)
    font_a = pygame.font.SysFont(None, 16)

    players = [None, None]  # two players limit
    active_players = {}

    colors = [
        create_surf((32, 32), (220, 180, 10)),
        create_surf((32, 32), (60, 230, 170)),
        create_surf((32, 32), (230, 20, 70)),
        create_surf((32, 32), (20, 170, 230)),
    ]
    are_no_controllers_connected = True

    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            elif event.type == pygame.JOYDEVICEADDED:
                if len(active_players) < len(players):
                    # connect controller
                    connect_joystick(
                        event.device_index, active_players, players, colors
                    )
                    are_no_controllers_connected = False
            elif event.type == pygame.JOYDEVICEREMOVED:
                # disconnect controller
                if event.instance_id in active_players:
                    disconnect_joystick(event.instance_id, active_players, players)
                    # check if there is at least one controller connected
                    are_no_controllers_connected = True
                    for player in players:
                        if player:
                            are_no_controllers_connected = False
                            break
            elif event.type == pygame.JOYBUTTONDOWN:
                if event.instance_id in active_players:
                    # join player
                    if event.button == 0:
                        index = active_players[event.instance_id]
                        players[index]["joined"] = True
                        print(f"P{index + 1} joined")
                    # leave player
                    if event.button == 1:
                        index = active_players[event.instance_id]
                        if players[index]["joined"]:
                            players[index]["joined"] = False
                            players[index]["pos"] = [WIDTH * 0.25 + index * 64, 128]
                            print(f"P{index + 1} leave")
            elif event.type == pygame.JOYAXISMOTION:
                if event.instance_id in active_players:
                    # change the color if player still hasn't joined
                    if event.axis == 0:
                        index = active_players[event.instance_id]
                        player = players[index]
                        if not player["joined"]:
                            if event.value >= 1.0:
                                player["surf_idx"] += 1
                            elif event.value <= -1.0:
                                player["surf_idx"] -= 1
                            player["surf_idx"] = player["surf_idx"] % len(colors)
                            player["surf"] = colors[player["surf_idx"]]

        screen.fill((30, 30, 30))
        pygame.draw.line(screen, (230, 230, 230), (0, 96), (WIDTH, 96), 2)

        # update and draw players
        for player in players:
            if player:
                control_player(player)
                screen.blit(player["surf"], player["pos"])

        # draw available colors
        for i, surf in enumerate(colors):
            screen.blit(surf, (WIDTH * 0.25 + i * 64, 32))

        # show message for connecting a controller
        if are_no_controllers_connected:
            screen.blit(
                font_b.render(
                    "Please connect a controller.",
                    True,
                    (230, 230, 230),
                    None,
                    500 - 20,
                ),
                (WIDTH * 0.3, HEIGHT * 0.5),
            )

        screen.blit(
            font_a.render(
                "A: join    B: leave    Joystick: move / change color",
                True,
                (230, 230, 230),
                None,
                WIDTH - 20,
            ),
            (10, HEIGHT - 20),
        )

        clock.tick(60)
        pygame.display.update()


if __name__ == "__main__":
    main()



from typing import Literal, Optional, Union
from typing_extensions import deprecated # added in 3.13

from pygame.bufferproxy import BufferProxy
from pygame.surface import Surface

from pygame.typing import FileLike, IntPoint, Point

_BufferStyle = Union[BufferProxy, bytes, bytearray, memoryview]
_to_bytes_format = Literal[
    "P", "RGB", "RGBX", "RGBA", "ARGB", "BGRA", "ABGR", "RGBA_PREMULT", "ARGB_PREMULT"
]
_from_buffer_format = Literal["P", "RGB", "BGR", "BGRA", "RGBX", "RGBA", "ARGB"]
_from_bytes_format = Literal["P", "RGB", "RGBX", "RGBA", "ARGB", "BGRA", "ABGR"]

def load(file: FileLike, namehint: str = "") -> Surface: ...
def load_sized_svg(file: FileLike, size: Point) -> Surface: ...
def save(surface: Surface, file: FileLike, namehint: str = "") -> None: ...
def get_sdl_image_version(linked: bool = True) -> Optional[tuple[int, int, int]]: ...
def get_extended() -> bool: ...
@deprecated("since 2.3.0. Use `pygame.image.tobytes` instead")
def tostring(
    surface: Surface,
    format: _to_bytes_format,
    flipped: bool = False,
    pitch: int = -1,
) -> bytes: ...
@deprecated("since 2.3.0. Use `pygame.image.frombytes` instead")
def fromstring(
    bytes: bytes,
    size: IntPoint,
    format: _from_bytes_format,
    flipped: bool = False,
    pitch: int = -1,
) -> Surface: ...
def tobytes(
    surface: Surface,
    format: _to_bytes_format,
    flipped: bool = False,
    pitch: int = -1,
) -> bytes: ...
def frombytes(
    bytes: bytes,
    size: IntPoint,
    format: _from_bytes_format,
    flipped: bool = False,
    pitch: int = -1,
) -> Surface: ...
def frombuffer(
    bytes: _BufferStyle,
    size: IntPoint,
    format: _from_buffer_format,
    pitch: int = -1,
) -> Surface: ...
def load_basic(file: FileLike, /) -> Surface: ...
def load_extended(file: FileLike, namehint: str = "") -> Surface: ...
def save_extended(surface: Surface, file: FileLike, namehint: str = "") -> None: ...



import webbrowser

# use relative import here and not absolute, so that `python -m docs` works at
# development time
from . import PKG_DIR


def open_docs():
    main_page = PKG_DIR / "index.html"
    if main_page.exists():
        url = main_page.as_uri()
    else:
        url = "https://pyga.me/docs/"
    webbrowser.open(url)


def main():
    print("Opening local documentation files in the browser.")
    print(
        "If you want to run a simple server instead, run 'python -m pygame.docs.serve'."
    )
    open_docs()


if __name__ == "__main__":
    main()



"""pygame.examples.go_over_there
This simple tech demo is showcasing the use of Vector2.move_towards()
using multiple circles to represent Vectors. Each circle will have a
random position and speed once the demo starts.

Mouse Controls:
* Use the mouse to click on a new target position

Keyboard Controls:
* Press R to restart the demo
"""

import pygame
import random

MIN_SPEED = 0.25
MAX_SPEED = 5
MAX_BALLS = 1600
SCREEN_SIZE = pygame.Vector2(1000, 600)
CIRCLE_RADIUS = 5

target_position = None
balls = []


class Ball:
    def __init__(self, position, speed):
        self.position = position
        self.speed = speed


def reset():
    global balls
    global target_position

    target_position = None
    balls = []
    for x in range(MAX_BALLS):
        pos = pygame.Vector2(
            random.randint(0, int(SCREEN_SIZE.x)), random.randint(0, int(SCREEN_SIZE.y))
        )
        speed = random.uniform(MIN_SPEED, MAX_SPEED)

        b = Ball(pos, speed)
        balls.append(b)


def main():
    global target_position
    global balls

    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE)
    clock = pygame.Clock()

    reset()
    delta_time = 0
    running = True
    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False

            if event.type == pygame.MOUSEBUTTONUP:
                target_position = pygame.mouse.get_pos()

            if event.type == pygame.KEYUP:
                if event.key == pygame.K_ESCAPE:
                    running = False

                if event.key == pygame.K_r:
                    reset()

        screen.fill((31, 143, 65))

        for o in balls:
            if target_position is not None:
                o.position.move_towards_ip(target_position, o.speed * delta_time)
            pygame.draw.circle(screen, (118, 207, 145), o.position, CIRCLE_RADIUS)

        pygame.display.flip()
        delta_time = clock.tick(60)
        pygame.display.set_caption(
            f"fps: {round(clock.get_fps(), 2)}, ball count: {len(balls)}"
        )

    pygame.quit()


if __name__ == "__main__":
    main()



"""Debug functionality that allows for more useful issue reporting"""

import platform
import sys
import traceback
import importlib
from collections.abc import Callable
from typing import Optional
from os import environ

from pygame.version import ver
from pygame.system import get_cpu_instruction_sets

ImportResult = tuple[str, bool, Optional[Callable]]


def str_from_tuple(version_tuple):
    """Converts a tuple like (2, 0, 20) into a string joined by periods

    Args:
        version_tuple: tuple(version_major, version_minor, version_patch)

    Returns:
       str: "major.minor.patch"
    """
    if version_tuple is None:
        return "None"

    strs = map(str, version_tuple)
    return ".".join(strs)


def attempt_import(module, function_name, output_str=""):
    """Attempts to import function_name from module

    Args:
        module: string representing module name
        function_name: string representing function name to be imported
        output_str: optional string to prepend error messages to if one occurs

    Returns:
        tuple(str, bool, Any):
            [0]: output_str + error_message
            [1]: True if successful, False if failed
            [2]: if successful, the thing that was imported, else None
    """
    try:
        mod = importlib.import_module(module)
        i = getattr(mod, function_name)
        success = True
    except (ImportError, AttributeError):
        i = None
        output_str += f"There was a problem with {module} import\n"
        output_str += "A default value will be returned for the version\n"
        output_str += traceback.format_exc() + "\n" + "=" * 20 + "\n"
        success = False

    return (output_str, success, i)


def _get_platform_info():
    """
    Internal helper to get platform information
    """
    cpu_inst_dict = get_cpu_instruction_sets()
    sse2 = 'Yes' if cpu_inst_dict['SSE2'] else 'No'
    avx2 = 'Yes' if cpu_inst_dict['AVX2'] else 'No'
    neon = 'Yes' if cpu_inst_dict['NEON'] else 'No'
    ret = f"Platform:\t\t{platform.platform()}\n"
    ret += f"System:\t\t\t{platform.system()}\n"
    ret += f"System Version:\t\t{platform.version()}\n"
    ret += f"Processor:\t\t{platform.processor()}\tSSE2: {sse2}\tAVX2: {avx2}\tNEON: {neon}\n"
    ret += (
        f"Architecture:\t\tBits: {platform.architecture()[0]}\t"
        f"Linkage: {platform.architecture()[1]}\n\n"
    )

    ret += f"Python:\t\t\t{platform.python_implementation()} {sys.version}\n"
    ret += f"pygame version:\t\t{ver}\n"
    return ret


def print_debug_info(filename=None):
    """Gets debug information for reporting bugs. Prints to console
    if filename is not specified, otherwise writes to that file
    (note: if filename is not an empty file, it will overwrite whatever is
    in there)

    Args:
        filename: string name of the file to save
    """
    debug_str = ""

    # keyword for compat with getters
    def default_return(linked=True):
        # pylint: disable=unused-argument
        return (-1, -1, -1)

    from pygame.display import (
        get_driver as get_display_driver,
        get_init as display_init,
    )
    from pygame.mixer import (
        get_driver as get_mixer_driver,
        get_init as mixer_init,
    )
    from pygame.base import get_sdl_version

    debug_str, *mixer = attempt_import(
        "pygame.mixer", "get_sdl_mixer_version", debug_str
    )
    if not mixer[0]:
        get_sdl_mixer_version = default_return
    else:
        get_sdl_mixer_version = mixer[1]

    debug_str, *font = attempt_import("pygame.font", "get_sdl_ttf_version", debug_str)
    if not font[0]:
        get_sdl_ttf_version = default_return
    else:
        get_sdl_ttf_version = font[1]

    debug_str, *image = attempt_import(
        "pygame.image", "get_sdl_image_version", debug_str
    )
    if not image[0]:
        get_sdl_image_version = default_return
    else:
        get_sdl_image_version = image[1]

    debug_str, *freetype = attempt_import("pygame.freetype", "get_version", debug_str)
    if not freetype[0]:
        ft_version = default_return
    else:
        ft_version = freetype[1]

    debug_str += _get_platform_info()

    debug_str += (
        f"SDL versions:\t\tLinked: {str_from_tuple(get_sdl_version())}\t"
        f"Compiled: {str_from_tuple(get_sdl_version(linked = False))}\n"
    )

    debug_str += (
        f"SDL Mixer versions:\tLinked: {str_from_tuple(get_sdl_mixer_version())}\t"
        f"Compiled: {str_from_tuple(get_sdl_mixer_version(linked = False))}\n"
    )

    debug_str += (
        f"SDL Font versions:\tLinked: {str_from_tuple(get_sdl_ttf_version())}\t"
        f"Compiled: {str_from_tuple(get_sdl_ttf_version(linked = False))}\n"
    )

    debug_str += (
        f"SDL Image versions:\tLinked: {str_from_tuple(get_sdl_image_version())}\t"
        f"Compiled: {str_from_tuple(get_sdl_image_version(linked = False))}\n"
    )

    debug_str += (
        f"Freetype versions:\tLinked: {str_from_tuple(ft_version())}\t"
        f"Compiled: {str_from_tuple(ft_version(linked = False))}\n\n"
    )

    if display_init():
        driver = get_display_driver()
        if driver.upper() != "X11":
            debug_str += f"Display Driver:\t\t{driver}\n"
        else:
            is_xwayland = (environ.get("XDG_SESSION_TYPE") == "wayland") or (
                "WAYLAND_DISPLAY" in environ
            )
            debug_str += f"Display Driver:\t\t{driver} ( xwayland == {is_xwayland} )\n"
    else:
        debug_str += "Display Driver:\t\tDisplay Not Initialized\n"

    if mixer_init():
        debug_str += f"Mixer Driver:\t\t{get_mixer_driver()}"
    else:
        debug_str += "Mixer Driver:\t\tMixer Not Initialized"

    if filename is None:
        print(debug_str)

    else:
        with open(filename, "w", encoding="utf8") as debugfile:
            debugfile.write(debug_str)



import unittest
import platform


class SysfontModuleTest(unittest.TestCase):
    def test_create_aliases(self):
        import pygame.sysfont

        pygame.sysfont.initsysfonts()
        pygame.sysfont.create_aliases()
        self.assertTrue(len(pygame.sysfont.Sysalias) > 0)

    def test_initsysfonts(self):
        import pygame.sysfont

        pygame.sysfont.initsysfonts()
        self.assertTrue(len(pygame.sysfont.get_fonts()) > 0)

    @unittest.skipIf("Darwin" not in platform.platform(), "Not mac we skip.")
    def test_initsysfonts_darwin(self):
        import pygame.sysfont

        self.assertTrue(len(pygame.sysfont.get_fonts()) > 10)

    def test_sysfont(self):
        import pygame.font

        pygame.font.init()
        arial = pygame.font.SysFont("Arial", 40)
        self.assertTrue(isinstance(arial, pygame.font.Font))

        pygame.font.SysFont(None, 40)

    @unittest.skipIf("Windows" not in platform.platform(), "Just for windows")
    def test_sysfont_settings(self):
        import pygame.font

        pygame.font.init()

        # The idea of this test is that we know we can expect Arial bold, Arial italic
        # Arial, and Arial italic bold to load different fonts on Windows. If that
        # isn't happening it may indicate a problem like
        # https://github.com/pygame-community/pygame-ce/issues/2677

        arial = pygame.font.SysFont("Arial", 40)
        arial_bold = pygame.font.SysFont("Arial", 40, bold=True)
        arial_italic = pygame.font.SysFont("Arial", 40, italic=True)
        arial_bold_italic = pygame.font.SysFont("Arial", 40, bold=True, italic=True)

        self.assertNotEqual(arial.style_name, arial_bold.style_name)
        self.assertNotEqual(arial.style_name, arial_italic.style_name)
        self.assertNotEqual(arial.style_name, arial_bold_italic.style_name)
        self.assertNotEqual(arial_bold.style_name, arial_italic.style_name)
        self.assertNotEqual(arial_italic.style_name, arial_bold_italic.style_name)
        self.assertNotEqual(arial_bold.style_name, arial_bold_italic.style_name)

    @unittest.skipIf(
        ("Darwin" in platform.platform() or "Windows" in platform.platform()),
        "Not unix we skip.",
    )
    def test_initsysfonts_unix(self):
        import pygame.sysfont

        self.assertTrue(len(pygame.sysfont.get_fonts()) > 0)

    @unittest.skipIf("Windows" not in platform.platform(), "Not windows we skip.")
    def test_initsysfonts_win32(self):
        import pygame.sysfont

        self.assertTrue(len(pygame.sysfont.get_fonts()) > 10)

    def test_sysfont_warnings(self):
        import pygame.font

        pygame.font.init()

        with self.assertWarns(UserWarning):
            pygame.font.SysFont("non-existent font", 40)

        with self.assertWarns(UserWarning):
            pygame.font.SysFont(bytes("non-existent font", "utf-8"), 40)

        with self.assertWarns(UserWarning):
            pygame.font.SysFont(("non-existent font", "non-existent font2"), 40)

        with self.assertWarns(UserWarning):
            pygame.font.SysFont(
                (bytes("non-existent font", "utf-8"), "non-existent font2"), 40
            )


###############################################################################

if __name__ == "__main__":
    unittest.main()



from typing import Optional, TypedDict

from pygame.typing import FileLike

class _MusicMetadataDict(TypedDict):
    title: str
    album: str
    artist: str
    copyright: str

def load(filename: FileLike, namehint: Optional[str] = "") -> None: ...
def unload() -> None: ...
def play(loops: int = 0, start: float = 0.0, fade_ms: int = 0) -> None: ...
def rewind() -> None: ...
def stop() -> None: ...
def pause() -> None: ...
def unpause() -> None: ...
def fadeout(time: int, /) -> None: ...
def set_volume(volume: float, /) -> None: ...
def get_volume() -> float: ...
def get_busy() -> bool: ...
def set_pos(pos: float, /) -> None: ...
def get_pos() -> int: ...
def queue(filename: FileLike, namehint: str = "", loops: int = 0) -> None: ...
def set_endevent(event_type: int, /) -> None: ...
def get_endevent() -> int: ...
def get_metadata(
    filename: Optional[FileLike] = None, namehint: str = ""
) -> _MusicMetadataDict: ...



import autoapi
import autoapi.documenters
from autoapi._objects import PythonClass


def build_signatures(object):
    name = object.short_name

    if isinstance(object, PythonClass):
        if object.constructor is not None:
            object = object.constructor
            object.obj["return_annotation"] = name
            object.obj["overloads"] = [
                (arg, name) for arg, _ in object.obj["overloads"]
            ]

        else:
            for child in object.children:
                if child.short_name == "__new__":
                    object = child
                    break

    if object is None or object.obj.get("args", None) is None:
        return

    sigs = [(object.obj["args"], object.obj["return_annotation"])]
    sigs.extend(object.obj["overloads"])

    for args, ret in sigs:
        arg_string = ""
        for modifier, arg_name, _, default in args:
            modifier = modifier or ""
            arg_name = arg_name or ""
            default = default or ""

            if default:
                default = "=" + default
            arg_string += f", {modifier}{arg_name}{default}"

        if arg_string:
            arg_string = arg_string[2:]

        if ret.count("[") > 2 or ret.count(",") > 3:
            ret = ret.split("[")[0]
            if ret in ("Optional", "Union"):
                ret = "..."

        yield f"| :sg:`{name}({arg_string}) -> {ret}`"


def get_doc(env, obj):
    if obj.docstring:
        return obj.docstring

    # If we don't already have docs, check if a python implementation exists of this
    # module and return its docstring if it does
    python_object = env.autoapi_all_objects.get(
        obj.id.replace("pygame", "src_py"), None
    )
    if python_object is not None:
        return python_object.docstring

    return ""


class AutopgDocumenter(autoapi.documenters.AutoapiDocumenter):
    def format_signature(self, **kwargs):
        return ""

    def get_doc(self, encoding=None, ignore=1):
        return [get_doc(self.env, self.object).splitlines()]

    def add_directive_header(self, sig: str) -> None:
        super().add_directive_header(sig)
        if "module" in self.objtype:
            sourcename = self.get_sourcename()
            self.add_line(f"   :synopsis: {self.get_doc()[0][0]}", sourcename)

    def get_object_members(self, want_all):
        members_check_module, members = super().get_object_members(want_all)
        members = (
            member
            for member in members
            if not member.object.obj.get("hide", False)
            and not member.object.imported
            and get_doc(self.env, member.object) != ""
        )

        return members_check_module, members

    def process_doc(self, docstrings):
        for docstring in docstrings:
            if not docstring:
                continue

            yield f"| :sl:`{docstring[0]}`"

            if "args" in self.object.obj or hasattr(self.object, "constructor"):
                yield from build_signatures(self.object)
            else:
                annotation = self.object.obj.get("annotation", None)
                if annotation is not None:
                    if annotation.count("[") > 2 or annotation.count(",") > 3:
                        annotation = "..."
                    yield f"| :sg:`{self.object.short_name} -> {annotation}`"

            yield from docstring[1:]

        yield ""


def setup(app):
    names = [
        "function",
        "property",
        "decorator",
        "class",
        "method",
        "data",
        "attribute",
        "module",
        "exception",
    ]

    for name in names:
        capitalized = name.capitalize()
        app.add_autodocumenter(
            type(
                f"Autopg{capitalized}Documenter",
                (
                    AutopgDocumenter,
                    getattr(autoapi.documenters, f"Autoapi{capitalized}Documenter"),
                ),
                {"objtype": f"pg{name}"},
            )
        )



from typing import TypedDict, Optional

# dict at runtime, this only exist for benefit of the typechecker
class _FingerDict(TypedDict):
    id: int
    x: float
    y: float
    pressure: float

def get_num_devices() -> int: ...
def get_device(index: int, /) -> int: ...
def get_num_fingers(device_id: int, /) -> int: ...
def get_finger(touchid: int, index: int) -> Optional[_FingerDict]: ...



from collections.abc import Callable
from typing import Union, Optional

ImportResult = tuple[str, bool, Optional[Callable]]

def str_from_tuple(version_tuple: Union[tuple[int, int, int], None]) -> str: ...
def attempt_import(module: str, function_name: str, output_str: str = "") -> ImportResult: ...
def print_debug_info(filename: Optional[str] = None) -> None: ...



from typing import Any, Optional, overload

from pygame.typing import _PathLike

def encode_string(
    obj: Optional[_PathLike],
    encoding: Optional[str] = "unicode_escape",
    errors: Optional[str] = "backslashreplace",
    etype: Optional[type[Exception]] = UnicodeEncodeError,
) -> bytes: ...
@overload
def encode_file_path(
    obj: Optional[_PathLike], etype: Optional[type[Exception]] = UnicodeEncodeError
) -> bytes: ...
@overload
def encode_file_path(
    obj: Any, etype: Optional[type[Exception]] = UnicodeEncodeError
) -> bytes: ...



# pygame-ce - Python Game Library
# Copyright (C) 2000-2003  Pete Shinners
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the Free
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Pete Shinners
# pete@shinners.org
"""sysfont, used in the font module to find system fonts"""

import warnings
import os
import sys
import itertools
import difflib
from os.path import basename, dirname, exists, join, splitext

from pygame.font import Font
from pygame import __file__ as pygame_main_file


OpenType_extensions = frozenset((".ttf", ".ttc", ".otf"))
Sysfonts = {}
Sysalias = {}

is_init = False


def _simplename(name):
    """create simple version of the font name"""
    # return alphanumeric characters of a string (converted to lowercase)
    return "".join(c.lower() for c in name if c.isalnum())


def _addfont(name, bold, italic, font, fontdict):
    """insert a font and style into the font dictionary"""
    if name not in fontdict:
        fontdict[name] = {}
    fontdict[name][bold, italic] = font


def initsysfonts_win32():
    """initialize fonts dictionary on Windows"""
    import winreg as _winreg

    fontdir = join(os.environ.get("WINDIR", "C:\\Windows"), "Fonts")
    fonts = {}

    # add fonts entered in the registry
    microsoft_font_dirs = [
        "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Fonts",
        "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Fonts",
    ]

    for domain in [_winreg.HKEY_LOCAL_MACHINE, _winreg.HKEY_CURRENT_USER]:
        for font_dir in microsoft_font_dirs:
            try:
                key = _winreg.OpenKey(domain, font_dir)
            except FileNotFoundError:
                continue

            for i in range(_winreg.QueryInfoKey(key)[1]):
                try:
                    # name is the font's name e.g. Times New Roman (TrueType)
                    # font is the font's filename e.g. times.ttf
                    name, font, _ = _winreg.EnumValue(key, i)
                except OSError:
                    break

                if splitext(font)[1].lower() not in OpenType_extensions:
                    continue
                if not dirname(font):
                    font = join(fontdir, font)

                # Some are named A & B, both names should be processed separately
                # Ex: the main Cambria file is marked as "Cambria & Cambria Math"
                for name in name.split("&"):
                    if os.path.exists(font):  # check if the file actually exists
                        _parse_font_entry_win(name, font, fonts)

    return fonts


def _parse_font_entry_win(name, font, fonts):
    """
    Parse out a simpler name and the font style from the initial file name.

    :param name: The font name
    :param font: The font file path
    :param fonts: The pygame font dictionary

    :return: Tuple of (bold, italic, name)
    """
    true_type_suffix = "(TrueType)"
    mods = ("demibold", "narrow", "light", "unicode", "bt", "mt")
    if name.endswith(true_type_suffix):
        name = name.rstrip(true_type_suffix).rstrip()
    name = name.lower().split()
    bold = italic = False
    for mod in mods:
        if mod in name:
            name.remove(mod)
    if "bold" in name:
        name.remove("bold")
        bold = True
    if "italic" in name:
        name.remove("italic")
        italic = True
    name = "".join(name)
    name = _simplename(name)

    _addfont(name, bold, italic, font, fonts)


def _parse_font_entry_darwin(name, filepath, fonts):
    """
    Parses a font entry for macOS

    :param name: The filepath without extensions or directories
    :param filepath: The full path to the font
    :param fonts: The pygame font dictionary to add the parsed font data to.
    """

    name = _simplename(name)

    mods = ("regular",)

    for mod in mods:
        if mod in name:
            name = name.replace(mod, "")

    bold = italic = False
    if "bold" in name:
        name = name.replace("bold", "")
        bold = True
    if "italic" in name:
        name = name.replace("italic", "")
        italic = True

    _addfont(name, bold, italic, filepath, fonts)


def _font_finder_darwin():
    locations = [
        "/Library/Fonts",
        "/Network/Library/Fonts",
        "/System/Library/Fonts",
        "/System/Library/Fonts/Supplemental",
    ]

    if username := os.getenv("USER"):
        locations.append(f"/Users/{username}/Library/Fonts")

    strange_root = "/System/Library/Assets/com_apple_MobileAsset_Font3"
    if exists(strange_root):
        locations += [
            f"{strange_root}/{loc}/AssetData" for loc in os.listdir(strange_root)
        ]

    fonts = {}

    for location in locations:
        if not exists(location):
            continue

        files = os.listdir(location)
        for file in files:
            name, extension = splitext(file)
            if extension in OpenType_extensions:
                _parse_font_entry_darwin(name, join(location, file), fonts)

    return fonts


def initsysfonts_darwin():
    """Read the fonts on macOS, and OS X."""
    # if the X11 binary exists... try to use that.
    #  Not likely to be there on pre 10.4.x ... or macOS 10.10+
    if exists("/usr/X11/bin/fc-list"):
        fonts = initsysfonts_unix("/usr/X11/bin/fc-list")
    # This fc-list path will work with the X11 from the OS X 10.3 installation
    # disc
    elif exists("/usr/X11R6/bin/fc-list"):
        fonts = initsysfonts_unix("/usr/X11R6/bin/fc-list")
    else:
        # eventually this should probably be the preferred solution
        fonts = _font_finder_darwin()

    return fonts


# read the fonts on unix
def initsysfonts_unix(path="fc-list"):
    """use the fc-list from fontconfig to get a list of fonts"""
    fonts = {}

    if sys.platform == "emscripten":
        return fonts

    import subprocess

    try:
        proc = subprocess.run(
            [path, ":", "file", "family", "style"],
            stdout=subprocess.PIPE,  # capture stdout
            stderr=subprocess.PIPE,  # capture stderr
            check=True,  # so that errors raise python exception which is handled below
            timeout=1,  # so that we don't hang the program waiting
        )

    except FileNotFoundError:
        warnings.warn(
            f"'{path}' is missing, system fonts cannot be loaded on your platform"
        )

    except subprocess.TimeoutExpired:
        warnings.warn(
            f"Process running '{path}' timed-out! System fonts cannot be loaded on "
            "your platform"
        )

    except subprocess.CalledProcessError as e:
        warnings.warn(
            f"'{path}' failed with error code {e.returncode}! System fonts cannot be "
            f"loaded on your platform. Error log is:\n{e.stderr}"
        )

    else:
        for entry in proc.stdout.decode("ascii", "ignore").splitlines():
            try:
                _parse_font_entry_unix(entry, fonts)
            except ValueError:
                # try the next one.
                pass

    return fonts


def _parse_font_entry_unix(entry, fonts):
    """
    Parses an entry in the unix font data to add to the pygame font
    dictionary.

    :param entry: An entry from the unix font list.
    :param fonts: The pygame font dictionary to add the parsed font data to.

    """
    filename, family, style = entry.split(":", 2)
    if splitext(filename)[1].lower() in OpenType_extensions:
        bold = "Bold" in style
        italic = "Italic" in style
        oblique = "Oblique" in style
        for name in family.strip().split(","):
            if name:
                break
        else:
            name = splitext(basename(filename))[0]

        _addfont(_simplename(name), bold, italic or oblique, filename, fonts)


def create_aliases():
    """Map common fonts that are absent from the system to similar fonts
    that are installed in the system
    """
    alias_groups = (
        (
            "monospace",
            "misc-fixed",
            "courier",
            "couriernew",
            "console",
            "fixed",
            "mono",
            "freemono",
            "bitstreamverasansmono",
            "verasansmono",
            "monotype",
            "lucidaconsole",
            "consolas",
            "dejavusansmono",
            "liberationmono",
        ),
        (
            "sans",
            "arial",
            "helvetica",
            "swiss",
            "freesans",
            "bitstreamverasans",
            "verasans",
            "verdana",
            "tahoma",
            "calibri",
            "gillsans",
            "segoeui",
            "trebuchetms",
            "ubuntu",
            "dejavusans",
            "liberationsans",
        ),
        (
            "serif",
            "times",
            "freeserif",
            "bitstreamveraserif",
            "roman",
            "timesroman",
            "timesnewroman",
            "dutch",
            "veraserif",
            "georgia",
            "cambria",
            "constantia",
            "dejavuserif",
            "liberationserif",
        ),
        ("wingdings", "wingbats"),
        ("comicsansms", "comicsans"),
    )
    for alias_set in alias_groups:
        for name in alias_set:
            if name in Sysfonts:
                found = Sysfonts[name]
                break
        else:
            continue
        for name in alias_set:
            if name not in Sysfonts:
                Sysalias[name] = found


def initsysfonts():
    """
    Initialise the sysfont module, called once. Locates the installed fonts
    and creates some aliases for common font categories.

    Has different initialisation functions for different platforms.
    """
    global is_init
    if is_init:
        # no need to re-init
        return

    if sys.platform == "win32":
        fonts = initsysfonts_win32()
    elif sys.platform == "darwin":
        fonts = initsysfonts_darwin()
    else:
        fonts = initsysfonts_unix()

    # Try to add the default font to sys fonts
    pygame_folder = os.path.dirname(os.path.abspath(pygame_main_file))
    default_font_path = os.path.join(pygame_folder, "freesansbold.ttf")
    _addfont("freesansbold", True, False, default_font_path, fonts)

    Sysfonts.update(fonts)
    create_aliases()
    is_init = True


def font_constructor(fontpath, size, bold, italic):
    """
    pygame.font specific declarations

    :param fontpath: path to a font.
    :param size: size of a font.
    :param bold: bold style, True or False.
    :param italic: italic style, True or False.

    :return: A font.Font object.
    """

    font = Font(fontpath, size)
    if bold:
        font.set_bold(True)
    if italic:
        font.set_italic(True)

    return font


def _load_single_font(name, bold=False, italic=False):
    fontname = None
    gotbold = False
    gotitalic = False
    single_name = _simplename(name)
    styles = Sysfonts.get(single_name)
    if not styles:
        styles = Sysalias.get(single_name)
    if styles:
        plainname = styles.get((False, False))
        fontname = styles.get((bold, italic))
        if not (fontname or plainname):
            # Neither requested style, nor plain font exists, so
            # return a font with the name requested, but an
            # arbitrary style.
            (style, fontname) = list(styles.items())[0]
            # Attempt to style it as requested. This can't
            # unbold or unitalicize anything, but it can
            # fake bold and/or fake italicize.
            if bold and style[0]:
                gotbold = True
            if italic and style[1]:
                gotitalic = True
        elif not fontname:
            fontname = plainname
        elif plainname != fontname:
            gotbold = bold
            gotitalic = italic
    return fontname, gotbold, gotitalic


# the exported functions


def SysFont(name, size, bold=False, italic=False, constructor=None):
    """pygame.font.SysFont(name, size, bold=False, italic=False, constructor=None) -> Font
    Create a pygame Font from system font resources.

    This will search the system fonts for the given font
    name. You can also enable bold or italic styles, and
    the appropriate system font will be selected if available.

    This will always return a valid Font object, and will
    fall back on the builtin pygame font if the given font
    is not found.

    Name can also be an iterable of font names, a string of
    comma-separated font names, or a bytes of comma-separated
    font names, in which case the set of names will be searched
    in order. Pygame uses a small set of common font aliases. If the
    specific font you ask for is not available, a reasonable
    alternative may be used.

    If optional constructor is provided, it must be a function with
    signature constructor(fontpath, size, bold, italic) which returns
    a Font instance. If None, a pygame.font.Font object is created.
    """
    if constructor is None:
        constructor = font_constructor

    initsysfonts()

    gotbold = gotitalic = False
    if name:
        if isinstance(name, (str, bytes)):
            name = name.split(b"," if isinstance(name, bytes) else ",")
        else:
            name = list(name)
        for idx, single_name in enumerate(name):
            if isinstance(single_name, bytes):
                name[idx] = single_name.decode()

        for single_name in name:
            fontname, gotbold, gotitalic = _load_single_font(single_name, bold, italic)
            if fontname:
                break

        else:
            if len(name) > 1:
                names = "', '".join(name)
                warnings.warn(
                    "None of the specified system fonts "
                    f"('{names}') could be found. "
                    "Using the default font instead."
                )
            else:
                # Identifies the closest matches to the font provided by
                # the user from the list of available fonts.
                matches = difflib.get_close_matches(
                    _simplename(name[0]), itertools.chain(Sysfonts, Sysalias), 3
                )
                if matches:
                    match_text = "Did you mean: '" + "', '".join(matches) + "'? "
                else:
                    match_text = ""

                warnings.warn(
                    f"The system font '{name[0]}' couldn't be "
                    f"found. {match_text}"
                    "Using the default font instead."
                )
    else:
        fontname = None

    set_bold = set_italic = False
    if bold and not gotbold:
        set_bold = True
    if italic and not gotitalic:
        set_italic = True

    return constructor(fontname, size, set_bold, set_italic)


def get_fonts():
    """pygame.font.get_fonts() -> list
    get a list of system font names

    Returns the list of all found system fonts. Note that
    the names of the fonts will be all lowercase with spaces
    removed. This is how pygame internally stores the font
    names for matching.
    """
    initsysfonts()
    return list(Sysfonts)


def match_font(name, bold=False, italic=False):
    """pygame.font.match_font(name, bold=False, italic=False) -> name
    find the filename for the named system font

    This performs the same font search as the SysFont()
    function, only it returns the path to the TTF file
    that would be loaded. The font name can also be an
    iterable of font names or a string/bytes of comma-separated
    font names to try.

    If no match is found, None is returned.
    """
    initsysfonts()

    fontname = None
    if isinstance(name, (str, bytes)):
        name = name.split(b"," if isinstance(name, bytes) else ",")

    for single_name in name:
        if isinstance(single_name, bytes):
            single_name = single_name.decode()

        single_name = _simplename(single_name)
        styles = Sysfonts.get(single_name)
        if not styles:
            styles = Sysalias.get(single_name)
        if styles:
            while not fontname:
                fontname = styles.get((bold, italic))
                if italic:
                    italic = 0
                elif bold:
                    bold = 0
                elif not fontname:
                    fontname = list(styles.values())[0]

        if fontname:
            break

    return fontname



"""
A simple script that reads the pyproject.toml at the root of the project, and
prints the project version specified in it.
"""

import sys
import pathlib

base_dir = pathlib.Path(__file__).parents[1]
config_file = base_dir / "pyproject.toml"
config_text = config_file.read_text()

if sys.version_info >= (3, 11):
    # more elegant and robust solution
    import tomllib

    conf = tomllib.loads(config_text)
    version = conf["project"]["version"]

else:
    # hacky solution, because we don't have tomllib in stdlib on older python
    # versions
    import re
    import ast

    finds = re.findall(r"^version\s*=(.*)", config_text, flags=re.MULTILINE)
    if len(finds) != 1:
        raise RuntimeError(f"Could not find precise version from {config_file}")

    version = ast.literal_eval(finds[0].strip())


_splits = version.split(".")

# handle optional dev tag
if len(_splits) == 3:
    _splits.append('""')
elif len(_splits) == 4:
    _splits[3] = f'".{_splits[3]}"'
else:
    raise ValueError("Invalid version!")

version_short = ".".join(_splits[:3])
version_macros = tuple(
    zip(
        ("PG_MAJOR_VERSION", "PG_MINOR_VERSION", "PG_PATCH_VERSION", "PG_VERSION_TAG"),
        _splits,
    )
)


if __name__ == "__main__":
    print(
        "\n".join(f"-D{key}={value}" for key, value in version_macros)
        if "--macros" in sys.argv
        else version
    )
